from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from psyke.extraction.cart.predictor import CartPredictor, LeafConstraints, LeafSequence
from psyke import get_default_random_seed, PedagogicalExtractor
from psyke.schema import GreaterThan, DiscreteFeature
from psyke.utils.logic import create_variable_list, create_head, create_term
from tuprolog.core import clause, Var, Struct
from tuprolog.theory import Theory, mutable_theory
from typing import Iterable
import pandas as pd


TREE_SEED = get_default_random_seed()


class Cart(PedagogicalExtractor):

    def __init__(self, predictor, max_depth: int = 3, max_leaves: int = None,
                 discretization: Iterable[DiscreteFeature] = None,
                 normalization=None, simplify: bool = True):
        super().__init__(predictor, discretization, normalization)
        self._cart_predictor = CartPredictor(normalization=normalization)
        self.depth = max_depth
        self.leaves = max_leaves
        self._simplify = simplify

    def _create_body(self, variables: dict[str, Var], constraints: LeafConstraints) -> Iterable[Struct]:
        results = []
        for feature_name, constraint, value in constraints:
            features = [d for d in self.discretization if feature_name in d.admissible_values]
            feature: DiscreteFeature = features[0] if len(features) > 0 else None
            results.append(create_term(variables[feature_name], constraint) if feature is None else
                           create_term(variables[feature.name],
                                       feature.admissible_values[feature_name],
                                       isinstance(constraint, GreaterThan)))
        return results

    @staticmethod
    def _simplify_nodes(nodes: list) -> Iterable:
        simplified = [nodes.pop(0)]
        while len(nodes) > 0:
            first_node = nodes[0][0]
            for condition in first_node:
                if all([condition in [node[0] for node in nodes][i] for i in range(len(nodes))]):
                    [node[0].remove(condition) for node in nodes]
            simplified.append(nodes.pop(0))
        return simplified

    def _create_theory(self, data: pd.DataFrame, mapping: dict[str: int], sort: bool = True) -> Theory:
        new_theory = mutable_theory()
        nodes = [node for node in self._cart_predictor]
        nodes = Cart._simplify_nodes(nodes) if self._simplify else nodes
        for (constraints, prediction) in nodes:
            if self.normalization is not None:
                m, s = self.normalization[data.columns[-1]]
                prediction = prediction * s + m
            if mapping is not None and prediction in mapping.values():
                for k, v in mapping.items():
                    if v == prediction:
                        prediction = k
                        break
            variables = create_variable_list(self.discretization, data, sort)
            new_theory.assertZ(
                clause(
                    create_head(data.columns[-1], list(variables.values()), prediction),
                    self._create_body(variables, constraints)
                )
            )
        return new_theory

    def _extract(self, data: pd.DataFrame, mapping: dict[str: int] = None, sort: bool = True) -> Theory:
        self._cart_predictor.predictor = DecisionTreeClassifier(random_state=TREE_SEED) \
            if isinstance(data.iloc[0, -1], str) or mapping is not None else DecisionTreeRegressor(random_state=TREE_SEED)
        if mapping is not None:
            data.iloc[:, -1] = data.iloc[:, -1].apply(lambda x: mapping[x] if x in mapping.keys() else x)
        self._cart_predictor.predictor.max_depth = self.depth
        self._cart_predictor.predictor.max_leaf_nodes = self.leaves
        self._cart_predictor.predictor.fit(data.iloc[:, :-1], data.iloc[:, -1])
        return self._create_theory(data, mapping, sort)

    def _predict(self, data) -> Iterable:
        return self._cart_predictor.predict(data)

    @property
    def n_rules(self) -> int:
        return self._cart_predictor.n_leaves



from enum import Enum
from math import log10
from random import Random

_DEFAULT_RANDOM_SEED: int = 123

ONNX_EXTENSION: str = '.onnx'

_random_options = dict(_deterministic_mode=True, _default_random_seed=_DEFAULT_RANDOM_SEED)

_random_seed_generator: Random = Random(_DEFAULT_RANDOM_SEED)

_DEFAULT_PRECISION: float = 1e-6

_precision_options: dict = {'precision': _DEFAULT_PRECISION}


class TypeNotAllowedException(Exception):

    def __init__(self, type_name: str):
        super().__init__('Type "' + type_name + '" not allowed for discretization.')


class Range:
    def __init__(self, mean: float, std: float):
        self.mean = mean
        self.std = std
        self.lower = mean
        self.upper = mean

    def left_infinite(self):
        self.lower = float('-inf')

    def right_infinite(self):
        self.upper = float('inf')

    def expand_left(self):
        self.lower -= self.std

    def expand_right(self):
        self.upper += self.std


def is_deterministic_mode():
    return _random_options['_deterministic_mode']


def set_deterministic_mode(value: bool):
    _random_options['_deterministic_mode'] = value


def get_default_random_seed():
    if is_deterministic_mode():
        return _random_options['_default_random_seed']
    else:
        return _random_seed_generator.randint(0, 1 << 64)


def set_default_random_seed(value: int):
    _random_options['_default_random_seed'] = value


def get_default_precision() -> float:
    return _precision_options['precision']


def get_int_precision() -> int:
    return -1 * int(log10(get_default_precision()))


def set_default_precision(value: float):
    _precision_options['precision'] = value


class Target(Enum):
    CLASSIFICATION = 1,
    CONSTANT = 2,
    REGRESSION = 3



from enum import Enum

import numpy as np
import pandas as pd

from psyke import Extractor
from psyke.tuning import Objective, Optimizer
from psyke.utils import Target


class CRASH(Optimizer):
    class Algorithm(Enum):
        ExACT = 1,
        CREAM = 2

    def __init__(self, predictor, dataframe: pd.DataFrame, max_mae_increase: float = 1.2,
                 min_rule_decrease: float = 0.9, readability_tradeoff: float = 0.1, max_depth: int = 10,
                 patience: int = 5, algorithm: Algorithm = Algorithm.CREAM, output: Target = Target.CONSTANT,
                 objective: Objective = Objective.MODEL, normalization=None):
        super().__init__(predictor, algorithm, dataframe, max_mae_increase, min_rule_decrease, readability_tradeoff,
                         max_depth, patience, objective, normalization)
        self.output = output

    def search(self):
        self.params = self.__search_depth()

    def __search_depth(self):
        params = []
        best = None

        for depth in range(1, self.max_depth + 1):
            p = self.__search_threshold(depth)
            b = Optimizer._best(p)[1]
            print()
            improvement = self._depth_improvement(
                [best[0], best[1]], [b[0], b[1]]
            ) if best is not None else np.inf

            best = b
            params += p

            if len(params) > 1 and improvement < 1.2:
                break
        return params

    def __search_threshold(self, depth):
        step = self.model_mae / 2.0
        threshold = self.model_mae * 0.9
        params = []
        patience = self.patience
        while patience > 0:
            print(f"{self.algorithm}. Depth: {depth}. Threshold = {threshold:.2f}. ", end="")
            extractor = Extractor.creepy(
                self.predictor, depth, threshold, self.output, 10, normalization=self.normalization,
                clustering=Extractor.cream if self.algorithm == CRASH.Algorithm.CREAM else Extractor.exact
            )
            _ = extractor.extract(self.dataframe)
            mae, n = (extractor.mae(self.dataframe, self.predictor) if self.objective == Objective.MODEL else
                      extractor.mae(self.dataframe)), extractor.n_rules
            print(f"MAE = {mae:.2f}, {n} rules")

            if len(params) == 0:
                params.append((mae, n, depth, threshold))
                threshold += step
                continue

            if (n == 1) or (mae == 0.0):
                params.append((mae, n, depth, threshold))
                break

            if mae > params[0][0] * self.max_mae_increase:
                break

            improvement = (params[-1][0] / mae) + (1 - n / params[-1][1])

            if improvement <= 1 or n > np.ceil(params[-1][1] * self.min_rule_decrease):
                patience -= 1
                step = max(step, abs(mae - threshold) / max(patience, 1))
            if mae != params[-1][0] or n != params[-1][1]:
                params.append((mae, n, depth, threshold))
            threshold += step
        return params

    def _print_params(self, name, params):
        print("**********************")
        print(f"Best {name}")
        print("**********************")
        print(f"MAE = {params[0]:.2f}, {params[1]} rules")
        print(f"Threshold = {params[3]:.2f}")
        print(f"Depth = {params[2]}")



from functools import partial

from kivy.uix.relativelayout import RelativeLayout
from sklearn.base import ClassifierMixin, RegressorMixin

from psyke.gui.view import INFO_PREDICTOR_MESSAGE, PREDICTOR_MESSAGE, PREDICTOR_PERFORMANCE_PREFIX, \
    INFO_PREDICTOR_PREFIX
from psyke.gui.view.layout import PanelBoxLayout, create_param_layout
from psyke.gui.model import PREDICTORS, FIXED_PREDICTOR_PARAMS
from psyke.utils.metrics import accuracy, f1, mae, mse, r2


class PredictorPanel(PanelBoxLayout):

    def __init__(self, controller, ratio, **kwargs):
        super().__init__(controller, 'Train', INFO_PREDICTOR_MESSAGE, 1, ratio,
                         PREDICTOR_MESSAGE, PREDICTORS, controller.set_predictor_param, **kwargs)

        self.parameter_panel = RelativeLayout(size_hint=(1, .83))

        self.add_widget(self.main_panel)
        self.add_widget(self.parameter_panel)
        self.add_widget(self.info_label)

    def select(self, spinner, text):
        if text == PREDICTOR_MESSAGE:
            self.controller.reset_predictor()
        else:
            self.controller.select_predictor(text)
            self.go_button.disabled = False
            params = PREDICTORS[text][1]
            self.parameter_panel.clear_widgets()
            for i, (name, (default, type, constraint)) in enumerate(dict(FIXED_PREDICTOR_PARAMS, **params).items()):
                if constraint is not None and constraint != self.controller.get_task_from_model():
                    continue
                self.parameter_panel.add_widget(create_param_layout(
                    name, default, type, partial(self.set_param, name), i, self.ratio))

    def go_action(self, button):
        self.controller.train_predictor()

    def set_info(self):
        predictor_name, predictor, predictor_params = self.controller.get_predictor_from_model()
        if predictor is None:
            self.info_label.text = INFO_PREDICTOR_MESSAGE
        else:
            self.info_label.text = ''
            for name, (_, _, constraint) in FIXED_PREDICTOR_PARAMS.items():
                if constraint is not None and constraint != self.controller.get_task_from_model():
                    continue
                self.info_label.text += f'{name} = {predictor_params[name]}\n'
            self.info_label.text += f'\n{INFO_PREDICTOR_PREFIX}Predictor: {predictor_name}\n'
            for name, (_, _, constraint) in PREDICTORS[predictor_name][1].items():
                if constraint is not None and constraint != self.controller.get_task_from_model():
                    continue
                self.info_label.text += f'{name} = {predictor_params[name]}\n'

            self.info_label.text += '\n' + PREDICTOR_PERFORMANCE_PREFIX

            test, action, preprocessing = self.controller.get_test_set_from_model()
            true = test.iloc[:, -1]
            predicted = predictor.predict(test.iloc[:, :-1])
            if action == 'Scale' and preprocessing is not None:
                m, s = preprocessing[test.columns[-1]]
                true = true * s + m
                predicted = predicted * s + m
            if isinstance(predictor, ClassifierMixin):
                self.info_label.text += f'Accuracy: {accuracy(true, predicted):.2f}\n' \
                                        f'F1: {f1(true, predicted):.2f}'
            elif isinstance(predictor, RegressorMixin):
                self.info_label.text += f'MAE: {mae(true, predicted):.2f}\n' \
                                   f'MSE: {mse(true, predicted):.2f}\n' \
                                   f'R2: {r2(true, predicted):.2f}'



from psyke import logger
from parameterized import parameterized_class
from psyke.utils import get_default_precision
from test.psyke import initialize
import unittest


@parameterized_class(initialize('iter'))
class TestIter(unittest.TestCase):

    def test_extract(self):
        logger.info(self.expected_theory)
        logger.info(self.extracted_theory)
        self.assertTrue(self.expected_theory.equals(self.extracted_theory, False))

    def test_predict(self):
        if isinstance(self.extracted_test_y_from_theory[0], str):
            self.assertTrue(all(self.extracted_test_y_from_theory == self.extracted_test_y_from_extractor))
        else:
            self.assertTrue(max(abs(self.extracted_test_y_from_theory - self.extracted_test_y_from_extractor)) <
                            get_default_precision())


if __name__ == '__main__':
    unittest.main()



def main(width=1500, height=770, resizable=False):
    from kivy import Config
    from kivy.app import App
    from kivy.lang import Builder

    Config.set('graphics', 'resizable', 1 if resizable else 0)
    Config.set('graphics', 'width', width)
    Config.set('graphics', 'height', height)
    Config.set('graphics', 'minimum_width', '1800')
    Config.set('graphics', 'minimum_height', '950')

    Builder.load_file('view/style.kv')

    class PSyKEApp(App):
        import os
        from psyke.gui.controller.Controller import Controller
        from psyke.gui.model.Model import Model

        model = Model(os.getcwd())
        controller = Controller(model)

        def build(self):
            return self.controller.screen

    PSyKEApp().run()



from psyke.extraction.trepan import Node, Split
from test import get_dataset
import math
import pandas as pd
import unittest


class TestSplit(unittest.TestCase):

    dataset: pd.DataFrame = get_dataset('iris')
    n_examples = dataset.shape[0]
    all_node = Node(dataset, n_examples)
    setosa_40 = Node(dataset.iloc[10:70, :], n_examples)
    setosa_40_complementar = Node(dataset.iloc[:10, :].append(dataset.iloc[70:, :]), n_examples)
    versicolor_25 = Node(dataset.iloc[40:75, :], n_examples)
    versicolor_25_complementar = Node(dataset.iloc[75:110, :], n_examples)

    def test_priority(self):
        self.assertTrue(math.isclose(-40/60-50/90-100,
                                     Split(self.all_node, (self.setosa_40, self.setosa_40_complementar)).priority))
        self.assertTrue(math.isclose((25 / 35) * - 2 - 200 + 200,
                                     Split(self.all_node, (self.versicolor_25, self.versicolor_25_complementar))
                                     .priority))


if __name__ == '__main__':
    unittest.main()



from collections import OrderedDict
from functools import partial
from math import floor

from kivy.uix.button import Button
from kivy.uix.label import Label
from kivy.uix.relativelayout import RelativeLayout
from kivy.uix.spinner import Spinner

from psyke.gui.view import NO_DATASET, COLOR_MAPS
from psyke.gui.view.layout import FeatureSelectionBoxLayout


class FeaturePanel(RelativeLayout):

    def __init__(self, controller, **kwargs):
        super().__init__(**kwargs)
        self.controller = controller
        self.features = {}
        self.feature_panels = {}
        self.plot_features = {}

        self.top_label = Label(text='Feature selection', size_hint=(.15, .17), pos_hint={'x': .01, 'y': .76})
        self.top_button = Button(text='Plot', size_hint=(.1, .17), pos_hint={'x': .15, 'y': .76})
        self.spinner_options = Spinner(
            pos_hint={'center_x': .31, 'center_y': .76 + .17 / 2.0}, size_hint=(.1, .17)
        )
        self.save_button = Button(text='Save', size_hint=(.1, .17), pos_hint={'x': .37, 'y': .76})
        self.spinner_options.bind(text=self.select)
        self.top_button.bind(on_press=self.plot)
        self.save_button.bind(on_press=self.save)
        self.alert_label = Label(size_hint=(.3, .17), pos_hint={'x': .59, 'y': .76}, font_size=20, color=(1, 0, 0))

        self.add_widget(self.top_label)
        self.add_widget(self.top_button)
        self.add_widget(self.spinner_options)
        self.add_widget(self.save_button)
        self.add_widget(self.alert_label)

        self.feature_panel = RelativeLayout(size_hint=(1., .7), pos_hint={'x': 0., 'y': 0.})
        self.add_widget(self.feature_panel)

    def init(self):
        self.features = {}
        self.feature_panels = {}
        self.plot_features = {}
        self.top_button.disabled = True
        self.save_button.disabled = True
        self.spinner_options.text = 'cmap'
        self.spinner_options.disabled = True
        self.reset_alert()
        self.feature_panel.clear_widgets()
        self.feature_panel.add_widget(Label(text=NO_DATASET, pos_hint={'x': -.2, 'y': .25}))
        self.feature_panel.add_widget(Label())

    def select(self, widget, value):
        self.controller.select_colormap(value if value != 'cmap' else None)

    @staticmethod
    def __inverse_mapping(discretization, name):
        original = [feature for feature in discretization if name in feature.admissible_values]
        return original[0].name if len(original) > 0 else name

    @staticmethod
    def __remove_duplicates(list_with_duplicates):
        return list(OrderedDict.fromkeys(list_with_duplicates))

    def set_info(self):
        dataset, pruned_dataset, action, preprocessing = self.controller.get_data_from_model()
        if dataset is not None:
            rankings = {feature: score for feature, score in self.controller.get_data_rankings_from_model()}
            pruned_columns = dataset.columns if pruned_dataset is None else pruned_dataset.columns
            columns = dataset.columns
            if action == 'Discretize' and preprocessing is not None:
                pruned_columns = self.__remove_duplicates(
                    [self.__inverse_mapping(preprocessing, column) for column in pruned_columns]
                )
                columns = self.__remove_duplicates(
                    [self.__inverse_mapping(preprocessing, column) for column in dataset.columns]
                )
            self.feature_panel.clear_widgets()
            self.top_button.disabled = False
            self.save_button.disabled = False
            self.spinner_options.disabled = False
            task = self.controller.get_task_from_model()
            self.spinner_options.values = [name for name, cmap in COLOR_MAPS[task]] if task is not None else []
            for i, feature in enumerate(columns):
                self.features[feature] = None if feature not in pruned_columns else \
                    'O' if feature == pruned_columns[-1] else 'I'
                self.plot_features[feature] = feature == dataset.columns[-1]
                self.feature_panels[feature] = FeatureSelectionBoxLayout(
                    feature, self.features[feature], rankings[feature] if self.features[feature] == 'I' else None,
                    partial(self.set_feature, feature), partial(self.set_plot_feature, feature),
                    pos_hint={'x': .01 + .24 * floor(i / 6.), 'y': 1 - 1. / 7. * (i % 6 + 1)})
                self.feature_panel.add_widget(self.feature_panels[feature])
        else:
            self.init()

    def set_feature(self, feature, button):
        reset_alert = False
        # if deselecting a feature
        if self.features[feature] == button.text:
            # if there are other features of the same kind, OK---e.g., I
            if len([k for k, v in self.features.items() if v == button.text]) > 1:
                self.features[feature] = None
                self.feature_panels[feature].plot_button.state = 'normal'
                self.plot_features[feature] = False
                reset_alert = True
            else:
                self.set_alert(f'Cannot remove the only selected {"input" if button.text == "I" else "output"} feature')
                self.feature_panels[feature].buttons[button.text].state = 'down'
        # if selecting input feature
        elif button.text == 'I':
            # deselected -> I, OK
            if self.features[feature] is None:
                self.features[feature] = button.text
                reset_alert = True
            # O -> I, ERROR
            else:
                self.set_alert('Cannot remove the only selected output feature')
                self.feature_panels[feature].buttons['I'].state = 'normal'
                self.feature_panels[feature].buttons['O'].state = 'down'
        # if selecting output feature
        else:
            # current feature is the only input feature, ERROR
            if len([k for k, v in self.features.items() if v == 'I' and k != feature]) == 0:
                self.set_alert('Cannot remove the only selected input feature')
                self.feature_panels[feature].buttons['O'].state = 'normal'
                self.feature_panels[feature].buttons['I'].state = 'down'
            else:
                self.features[feature] = button.text
                to_normal = [k for k, v in self.features.items() if v == 'O' and k != feature]
                if len(to_normal) > 0:
                    self.feature_panels[to_normal[0]].buttons['O'].state = 'normal'
                    self.features[to_normal[0]] = None
                    self.feature_panels[to_normal[0]].plot_button.state = 'normal'
                    self.plot_features[to_normal[0]] = False
                self.feature_panels[feature].plot_button.state = 'down'
                self.plot_features[feature] = True
                reset_alert = True
        if reset_alert:
            self.reset_alert(self.controller.reload_dataset(self.features))

    def set_plot_feature(self, feature, button):
        if self.features[feature] is None:
            self.set_alert('Cannot plot non-selected features')
            self.feature_panels[feature].plot_button.state = 'normal'
        elif self.features[feature] == 'O':
            self.set_alert('Output feature must be plotted')
            self.feature_panels[feature].plot_button.state = 'down'
        else:
            inputs = [k for k, v in self.features.items() if v == 'I']
            active = len([k for k, v in self.plot_features.items() if v and k in inputs])
            if (self.plot_features[feature] and active == 2) or (not self.plot_features[feature] and active < 2):
                self.plot_features[feature] = not self.plot_features[feature]
                self.reset_alert()
            else:
                self.set_alert('Cannot plot less than 1 input feature' if self.plot_features[feature] else
                               'Cannot plot more than 2 input features')
                state = self.feature_panels[feature].plot_button.state
                self.feature_panels[feature].plot_button.state = 'down' if state == 'normal' else 'normal'

    def plot(self, button, save=False):
        inputs = len([k for k, v in self.features.items() if v == 'I' and self.plot_features[k]])
        task = self.controller.get_task_from_model()
        if task == 'Classification' and inputs < 2:
            self.set_alert('Classification plots require two input features')
        elif task == 'Regression' and inputs == 0:
            self.set_alert('Regression plots require at least one input feature')
        else:
            self.controller.plot(self.features, [k for k, v in self.plot_features.items() if v], save)
            self.reset_alert()

    def save(self, button, save=False):
        self.plot(button, True)

    def set_alert(self, text: str = ''):
        self.alert_label.text = text

    def reset_alert(self, reset=True):
        if reset:
            self.set_alert()
        rankings = self.controller.get_data_rankings_from_model()
        rankings = {feature: None for feature in self.features} if rankings is None else \
            {feature: score for feature, score in rankings}
        for feature, panel in self.feature_panels.items():
            panel.set_text(feature, rankings[feature] if self.features[feature] == 'I' else None)



from __future__ import annotations

import math
from hashlib import sha256
from typing import Iterable, List
import pandas as pd
from pandas.core.util.hashing import hash_pandas_object
from pandas.api.types import is_string_dtype, is_numeric_dtype, is_integer_dtype
from sklearn.preprocessing import StandardScaler
from sympy.core.containers import OrderedSet

from psyke import DiscreteFeature
from psyke.schema import LessThan, GreaterThan, Between, Value, Constant
from psyke.utils import TypeNotAllowedException, Range


def split_features(dataframe: pd.DataFrame) -> Iterable[DiscreteFeature]:
    result = []
    features = {'V' + str(index + 1): column for index, column in enumerate(dataframe.columns)}
    for feature, column in features.items():
        values = set(dataframe[column])
        result.append(DiscreteFeature(feature, {feature + '_' + str(i): v for i, v in enumerate(values)}))
    return result


def get_discrete_features_supervised(dataframe: pd.DataFrame) -> Iterable[DiscreteFeature]:
    result = OrderedSet()
    for feature in dataframe.columns[:-1]:
        result.add(DiscreteFeature(feature, create_set(feature, dataframe)))
    return result


def create_set(feature: str, dataframe: pd.DataFrame) -> dict[str, Value]:
    if is_string_dtype(dataframe[feature]) or is_integer_dtype(dataframe[feature]):
        values = dataframe[feature].unique()
    elif is_numeric_dtype(dataframe[feature]):
        values = create_ranges(feature, dataframe)
    else:
        raise TypeNotAllowedException(dataframe[feature].dtype)
    return {"{}_{}".format(feature, i): create_original_value(v) for (i, v) in enumerate(values)}


def create_original_value(value: Range | str | int) -> Value:
    if isinstance(value, Range):
        if value.lower == float('-inf'):
            return LessThan(value.upper)
        if value.upper == float('inf'):
            return GreaterThan(value.lower)
        return Between(value.lower, value.upper)
    return Constant(value)


def create_ranges(feature: str, dataframe: pd.DataFrame) -> Iterable[Range]:
    ranges = init_ranges(feature, dataframe)
    expand_ranges(ranges)
    ranges[0].left_infinite()
    ranges[-1].right_infinite()
    return ranges


def expand_ranges(ranges: Iterable[Range]):
    for r1, r2 in zip(ranges[0:-1], ranges[1:]):
        while r1.upper < r2.lower:
            r1.expand_right()
            r2.expand_left()
        mean = ((r1.upper - r1.std + r2.lower + r2.std) / 2)
        r1.upper = mean
        r2.lower = mean


def init_ranges(feature: str, dataframe: pd.DataFrame) -> Iterable[Range]:
    desc = [dataframe[dataframe.iloc[:, -1] == v].describe()[feature] for v in dataframe.iloc[:, -1].unique()]
    desc = [(d['mean'], d['std']) for d in desc]
    desc.sort()
    return [Range(d[0], d[1]) for d in desc]


def get_discrete_features_equal_frequency(
        dataframe: pd.DataFrame,
        bins: int = None,
        output: bool = True,
        bin_names: List[str] = []
) -> Iterable[DiscreteFeature]:
    features = dataframe.columns[:-1] if output else dataframe.columns
    result = set()
    if bins is None:
        if len(bin_names) > 0:
            bins = len(bin_names)
        else:
            raise ValueError("No bins nor bin_names have been provided")
    elif bins > 0:
        if len(bin_names) == 0:
            bin_names = range(0, bins)
        elif len(bin_names) == bins:
            pass
        else:
            raise ValueError("Mismatch among the provided amount of bins and the bin_names")
    else:
        raise ValueError("Negative amount of bins makes no sense")
    for feature in features:
        values = sorted(dataframe[feature])
        intervals = [values[i * math.ceil(len(values) / bins)] for i in range(1, bins)]
        starting_interval: list[Value] = [LessThan(intervals[0])]
        ending_interval: list[Value] = [GreaterThan(intervals[-1])]
        middle_intervals: list[Value] = [Between(intervals[i], intervals[i + 1]) for i in range(0, len(intervals) - 1)]
        new_intervals = starting_interval + middle_intervals + ending_interval
        new_feature_names = [feature + '_' + str(i) for i in range(0, bins)]
        new_features = {new_feature_names[i]: new_intervals[i] for i in range(0, bins)}
        result.add(DiscreteFeature(feature, new_features))
    return result


def get_discrete_dataset(dataset: pd.DataFrame, discrete_features: Iterable[DiscreteFeature],
                         sort: bool = True) -> pd.DataFrame:
    """
    Create a new dataset mapping the old features into the new discrete features.
    Note: some algorithms require the same SORTED feature to be replicable due to rule optimization and other stuffs.
    Therefore the new features are alphabetically sorted.
    This is not strictly necessary because internally those algorithms perform the sorting themself.
    However it is a good idea to have this same function returning the same result w.r.t. the inputs.

    :param dataset: the original dataset
    :param discrete_features: mapping for the features
    :param sort: alphabetically sort new features
    :return: the new discrete dataset
    """
    columns_name = [key for feature in discrete_features for key, _ in feature.admissible_values.items()]
    if sort:
        columns_name = sorted(columns_name)
    new_dataset = pd.DataFrame(columns=columns_name)
    for feature in discrete_features:
        for index, value in enumerate(dataset[feature.name]):
            for key, admissible_value in feature.admissible_values.items():
                new_dataset.loc[index, key] = int(admissible_value.is_in(value))

    for feature in discrete_features:
        for new_feature in feature.admissible_values.keys():
            new_dataset[new_feature] = new_dataset[new_feature].astype(str).astype(int)

    return new_dataset


def get_scaled_dataset(dataset: pd.DataFrame) -> tuple[pd.DataFrame, dict[str, tuple[float, float]]]:
    scaler = StandardScaler()
    scaler.fit(dataset)
    normalization = {key: (m, s) for key, m, s in zip(dataset.columns, scaler.mean_, scaler.scale_)}
    return pd.DataFrame(scaler.transform(dataset), columns=dataset.columns, index=dataset.index), normalization


def scale_dataset(dataset: pd.DataFrame, normalization: dict[str, tuple[float, float]]) -> pd.DataFrame:
    new_data = pd.DataFrame()
    for column in dataset.columns:
        m, s = normalization[column]
        new_data[column] = (dataset[column] - m) / s
    return new_data


class HashableDataFrame(pd.DataFrame):
    def __init__(self, obj):
        super().__init__(obj)

    def __hash__(self):
        hash_value = sha256(hash_pandas_object(self, index=True).values)
        hash_value = hash(hash_value.hexdigest())
        return hash_value

    def __eq__(self, other):
        return self.equals(other)



from __future__ import annotations
from abc import ABC
import numpy as np
import pandas as pd
from numpy import argmax
from sklearn.linear_model import LinearRegression
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score, f1_score, accuracy_score
from psyke.schema import DiscreteFeature
from psyke.utils import get_default_random_seed
from tuprolog.theory import Theory
from typing import Iterable
import logging


logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger('psyke')


class Extractor(object):
    """
    An explanator capable of extracting rules from trained black box.

    Parameters
    ----------
    predictor : the underling black box predictor.
    discretization : A collection of sets of discretised features.
        Each set corresponds to a set of features derived from a single non-discrete feature.
    """

    def __init__(self, predictor, discretization: Iterable[DiscreteFeature] = None, normalization=None):
        self.predictor = predictor
        self.discretization = [] if discretization is None else list(discretization)
        self.normalization = normalization

    def extract(self, dataframe: pd.DataFrame, mapping: dict[str: int] = None, sort: bool = True) -> Theory:
        """
        Extracts rules from the underlying predictor.

        :param dataframe: is the set of instances to be used for the extraction.
        :param mapping: for one-hot encoding.
        :param sort: alphabetically sort the variables of the head of the rules.
        :return: the theory created from the extracted rules.
        """
        raise NotImplementedError('extract')

    def predict(self, dataframe: pd.DataFrame, mapping: dict[str: int] = None) -> Iterable:
        """
        Predicts the output values of every sample in dataset.

        :param dataframe: is the set of instances to predict.
        :param mapping: for one-hot encoding.
        :return: a list of predictions.
        """
        ys = self._predict(dataframe)
        if mapping is not None:
            inverse_mapping = {v: k for k, v in mapping.items()}
            ys = [inverse_mapping[y] for y in ys]
        return ys

    def _predict(self, dataframe: pd.DataFrame) -> Iterable:
        raise NotImplementedError('predict')

    def unscale(self, values, name):
        if self.normalization is None or isinstance(values, LinearRegression):
            return values
        if isinstance(values, Iterable):
            idx = [value is not None for value in values]
            values[idx] = values[idx] * self.normalization[name][1] + self.normalization[name][0]
        else:
            values = values * self.normalization[name][1] + self.normalization[name][0]
        return values

    def regression_score(self, dataframe: pd.DataFrame, predictor=None, scoring_function=mean_absolute_error):
        predictions = np.array(self.predict(dataframe.iloc[:, :-1]))
        idx = [prediction is not None for prediction in predictions]
        true = self.unscale(dataframe.iloc[idx, -1] if predictor is None else
                            predictor.predict(dataframe.iloc[idx, :-1]).flatten(), dataframe.columns[-1])
        return scoring_function(true, self.unscale(predictions[idx], dataframe.columns[-1]))

    def mae(self, dataframe: pd.DataFrame, predictor=None) -> float:
        """
        Calculates the predictions' MAE w.r.t. the instances given as input.

        :param dataframe: is the set of instances to be used to calculate the mean absolute error.
        :param predictor: if provided, its predictions on the dataframe are taken instead of the dataframe instances.
        :return: the mean absolute error (MAE) of the predictions.
        """
        return self.regression_score(dataframe, predictor, mean_absolute_error)

    def mse(self, dataframe: pd.DataFrame, predictor=None) -> float:
        """
        Calculates the predictions' MSE w.r.t. the instances given as input.

        :param dataframe: is the set of instances to be used to calculate the mean squared error.
        :param predictor: if provided, its predictions on the dataframe are taken instead of the dataframe instances.
        :return: the mean squared error (MSE) of the predictions.
        """
        return self.regression_score(dataframe, predictor, mean_squared_error)

    def r2(self, dataframe: pd.DataFrame, predictor=None) -> float:
        """
        Calculates the predictions' R2 score w.r.t. the instances given as input.

        :param dataframe: is the set of instances to be used to calculate the R2 score.
        :param predictor: if provided, its predictions on the dataframe are taken instead of the dataframe instances.
        :return: the R2 score of the predictions.
        """
        return self.regression_score(dataframe, predictor, r2_score)

    def accuracy(self, dataframe: pd.DataFrame, predictor=None) -> float:
        """
        Calculates the predictions' accuracy classification score w.r.t. the instances given as input.

        :param dataframe: is the set of instances to be used to calculate the accuracy classification score.
        :param predictor: if provided, its predictions on the dataframe are taken instead of the dataframe instances.
        :return: the accuracy classification score of the predictions.
        """
        predictions = np.array(self.predict(dataframe.iloc[:, :-1]))
        idx = [prediction is not None for prediction in predictions]
        return accuracy_score(dataframe.iloc[idx, -1] if predictor is None else
                              predictor.predict(dataframe.iloc[idx, :-1]).flatten(),
                              predictions[idx])

    def f1(self, dataframe: pd.DataFrame, predictor=None) -> float:
        """
        Calculates the predictions' F1 score w.r.t. the instances given as input.

        :param dataframe: is the set of instances to be used to calculate the F1 score.
        :param predictor: if provided, its predictions on the dataframe are taken instead of the dataframe instances.
        :return: the F1 score of the predictions.
        """
        predictions = np.array(self.predict(dataframe.iloc[:, :-1])[:])
        idx = [prediction is not None for prediction in predictions]
        return f1_score(dataframe.iloc[idx, -1] if predictor is None else
                        predictor.predict(dataframe.iloc[idx, :-1]).flatten(),
                        predictions[idx], average='weighted')

    @staticmethod
    def exact(depth: int, error_threshold: float, output, gauss_components: int = 2):
        """
        Creates a new ExACT instance.
        """
        from psyke.clustering.exact import ExACT
        return ExACT(depth, error_threshold, output, gauss_components)

    @staticmethod
    def cream(depth: int, error_threshold: float, output, gauss_components: int = 2):
        """
        Creates a new CREAM instance.
        """
        from psyke.clustering.cream import CREAM
        return CREAM(depth, error_threshold, output, gauss_components)

    @staticmethod
    def cart(predictor, max_depth: int = 3, max_leaves: int = 3,
             discretization: Iterable[DiscreteFeature] = None, normalization=None, simplify: bool = True) -> Extractor:
        """
        Creates a new Cart extractor.
        """
        from psyke.extraction.cart import Cart
        return Cart(predictor, max_depth, max_leaves, discretization=discretization, normalization=normalization,
                    simplify=simplify)

    @staticmethod
    def iter(predictor, min_update: float = 0.1, n_points: int = 1, max_iterations: int = 600, min_examples: int = 250,
             threshold: float = 0.1, fill_gaps: bool = True, normalization: dict[str, tuple[float, float]] = None,
             output=None, seed: int = get_default_random_seed()) -> Extractor:
        """
        Creates a new ITER extractor.
        """
        from psyke.extraction.hypercubic.iter import ITER
        return ITER(predictor, min_update, n_points, max_iterations, min_examples, threshold, fill_gaps,
                    normalization, output, seed)

    @staticmethod
    def gridex(predictor, grid, min_examples: int = 250, threshold: float = 0.1,
               normalization: dict[str, tuple[float, float]] = None,
               seed: int = get_default_random_seed()) -> Extractor:
        """
        Creates a new GridEx extractor.
        """
        from psyke.extraction.hypercubic.gridex import GridEx
        return GridEx(predictor, grid, min_examples, threshold, normalization, seed)

    @staticmethod
    def gridrex(predictor, grid, min_examples: int = 250, threshold: float = 0.1,
                normalization: dict[str, tuple[float, float]] = None,
                seed: int = get_default_random_seed()) -> Extractor:
        """
        Creates a new GridREx extractor.
        """
        from psyke.extraction.hypercubic.gridrex import GridREx
        return GridREx(predictor, grid, min_examples, threshold, normalization, seed)

    @staticmethod
    def creepy(predictor, clustering, depth: int, error_threshold: float, output, gauss_components: int = 2,
               ranks: [(str, float)] = [], ignore_threshold: float = 0.0,
               normalization: dict[str, tuple[float, float]] = None) -> Extractor:
        """
        Creates a new CReEPy extractor.
        """
        from psyke.extraction.hypercubic.creepy import CReEPy
        return CReEPy(predictor, depth, error_threshold, output, gauss_components, ranks, ignore_threshold,
                      normalization, clustering)

    @staticmethod
    def real(predictor, discretization=None) -> Extractor:
        """
        Creates a new REAL extractor.
        """
        from psyke.extraction.real import REAL
        return REAL(predictor, [] if discretization is None else discretization)

    @staticmethod
    def trepan(predictor, discretization=None, min_examples: int = 0, max_depth: int = 3,
               split_logic=None) -> Extractor:
        """
        Creates a new Trepan extractor.
        """
        from psyke.extraction.trepan import Trepan, SplitLogic
        if split_logic is None:
            split_logic = SplitLogic.DEFAULT
        return Trepan(predictor, [] if discretization is None else discretization, min_examples, max_depth, split_logic)


class PedagogicalExtractor(Extractor, ABC):

    def extract(self, dataframe: pd.DataFrame, mapping: dict[str: int] = None, sort: bool = True) -> Theory:
        new_y = self.predictor.predict(dataframe.iloc[:, :-1])
        if mapping is not None:
            if hasattr(new_y[0], 'shape'):
                # One-hot encoding for multi-class tasks
                if len(new_y[0].shape) > 0 and new_y[0].shape[0] > 1:
                    new_y = [argmax(y, axis=0) for y in new_y]
                # One-hot encoding for binary class tasks
                else:
                    new_y = [round(y[0]) for y in new_y]
        new_y = pd.DataFrame(new_y).set_index(dataframe.index)
        data = dataframe.iloc[:, :-1].copy().join(new_y)
        data.columns = dataframe.columns
        return self._extract(data, mapping, sort)

    def _extract(self, dataframe: pd.DataFrame, mapping: dict[str: int] = None, sort: bool = True) -> Theory:
        raise NotImplementedError('predict')



from functools import partial

from kivy.uix.relativelayout import RelativeLayout
from sklearn.base import ClassifierMixin, RegressorMixin

from psyke.gui.model import EXTRACTORS
from psyke.gui.view.layout import PanelBoxLayout, create_param_layout
from psyke.gui.view import INFO_EXTRACTOR_MESSAGE, EXTRACTOR_MESSAGE, INFO_EXTRACTOR_PREFIX, \
    EXTRACTOR_PERFORMANCE_PREFIX
from psyke.utils.metrics import accuracy, f1, mae, mse, r2


class ExtractorPanel(PanelBoxLayout):

    def __init__(self, controller, ratio=1, **kwargs):
        super().__init__(controller, 'Extract', INFO_EXTRACTOR_MESSAGE, 1, ratio,
                         EXTRACTOR_MESSAGE, EXTRACTORS, controller.set_extractor_param, **kwargs)

        self.parameter_panel = RelativeLayout(size_hint=(1, .83))

        self.add_widget(self.main_panel)
        self.add_widget(self.parameter_panel)
        self.add_widget(self.info_label)

    def set_info(self):
        extractor_name, extractor, extractor_params = self.controller.get_extractor_from_model()
        if extractor is None:
            self.info_label.text = INFO_EXTRACTOR_MESSAGE
        else:
            self.info_label.text = f'\n{INFO_EXTRACTOR_PREFIX}Predictor: {extractor_name}\n'
            for name, (_, _, constraint) in EXTRACTORS[extractor_name][1].items():
                if constraint is not None and constraint != self.controller.get_task_from_model():
                    continue
                self.info_label.text += f'{name} = {extractor_params[name]}\n'

            self.info_label.text += '\n' + EXTRACTOR_PERFORMANCE_PREFIX
            self.info_label.text += f'N. rules: {extractor.n_rules}\n'

            test, action, preprocessing = self.controller.get_test_set_from_model()
            predictor = self.controller.get_predictor_from_model()[1]
            extracted = extractor.predict(test.iloc[:, :-1])
            true = test.iloc[:, -1]
            predicted = predictor.predict(test.iloc[:, :-1])
            if action == 'Scale' and preprocessing is not None:
                m, s = preprocessing[test.columns[-1]]
                extracted = extracted * s + m
                true = true * s + m
                predicted = predicted * s + m

            if isinstance(predictor, ClassifierMixin):
                labels = ['Acc.', 'F1']
                metrics = [accuracy, f1]
            elif isinstance(predictor, RegressorMixin):
                labels = ['MAE', 'MSE', 'R2']
                metrics = [mae, mse, r2]
            else:
                raise NotImplementedError

            for label, metric in zip(labels, metrics):
                self.info_label.text += \
                    f'{label}: {metric(true, extracted):.2f} (data), {metric(predicted, extracted):.2f} (BB)\n'

    def select(self, spinner, text):
        if text == EXTRACTOR_MESSAGE:
            self.controller.reset_extractor()
        else:
            self.controller.select_extractor(text)
            self.go_button.disabled = False
            params = EXTRACTORS[text][1]
            self.parameter_panel.clear_widgets()
            for i, (name, (default, type, constraint)) in enumerate(params.items()):
                if constraint is not None and constraint != self.controller.get_task_from_model():
                    continue
                self.parameter_panel.add_widget(create_param_layout(
                    name, default, type, partial(self.set_param, name), i, self.ratio))

    def go_action(self, button):
        self.controller.train_extractor()



import unittest
from sklearn.model_selection import train_test_split
from sklearn.tree import DecisionTreeRegressor
from psyke import Extractor, get_default_random_seed
from psyke.extraction.hypercubic import Grid
from test import get_dataset


class TestSimplifyFormatter(unittest.TestCase):

    def test_simplify_formatter(self):
        iris_data = get_dataset('house')
        train, test = train_test_split(iris_data, test_size=0.5, random_state=get_default_random_seed())
        predictor = DecisionTreeRegressor()
        predictor.fit(train.iloc[:, :-1], train.iloc[:, -1])
        extractor = Extractor.gridrex(predictor, Grid())
        theory = extractor.extract(train)
        # print(pretty_theory(theory))


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from random import Random
from typing import Iterable
import numpy as np
import pandas as pd
from sklearn.base import ClassifierMixin
from tuprolog.theory import Theory
from psyke import PedagogicalExtractor
from psyke.extraction.hypercubic import HyperCube, HyperCubeExtractor
from psyke.extraction.hypercubic.hypercube import GenericCube
from psyke.extraction.hypercubic.utils import MinUpdate, Expansion
from psyke.utils import get_default_random_seed, Target

DomainProperties = (Iterable[MinUpdate], GenericCube)


class ITER(PedagogicalExtractor, HyperCubeExtractor):
    """
    Explanator implementing ITER algorithm, doi:10.1007/11823728_26.
    """

    def __init__(self, predictor, min_update, n_points, max_iterations, min_examples, threshold, fill_gaps,
                 normalization, output: Target = Target.CONSTANT, seed=get_default_random_seed()):
        super().__init__(predictor, normalization)
        if output is Target.REGRESSION:
            raise NotImplementedError
        self.predictor = predictor
        self.min_update = min_update
        self.n_points = n_points
        self.max_iterations = max_iterations
        self.min_examples = min_examples
        self.threshold = threshold
        self.fill_gaps = fill_gaps
        self._output = Target.CLASSIFICATION if isinstance(predictor, ClassifierMixin) else \
            output if output is not None else Target.CONSTANT
        self.__generator = Random(seed)

    def _best_cube(self, dataframe: pd.DataFrame, cube: GenericCube, cubes: Iterable[Expansion]) -> Expansion | None:
        expansions = []
        for limit in cubes:
            count = limit.cube.count(dataframe)
            dataframe = dataframe.append(limit.cube.create_samples(self.min_examples - count, self.__generator))
            limit.cube.update(dataframe, self.predictor)
            expansions.append(Expansion(
                limit.cube, limit.feature, limit.direction,
                abs(cube.output - limit.cube.output) if self._output is Target.CONSTANT else
                1 - int(cube.output == limit.cube.output)
            ))
        if len(expansions) > 0:
            return sorted(expansions, key=lambda e: e.distance)[0]
        return None

    def _calculate_min_updates(self, surrounding: GenericCube) -> Iterable[MinUpdate]:
        return [MinUpdate(name, (interval[1] - interval[0]) * self.min_update) for (name, interval) in
                surrounding.dimensions.items()]

    @staticmethod
    def _create_range(cube: GenericCube, domain: DomainProperties, feature: str, direction: str)\
            -> tuple[GenericCube, tuple[float, float]]:
        min_updates, surrounding = domain
        a, b = cube[feature]
        size = [min_update for min_update in min_updates if min_update.name == feature][0].value
        return (cube.copy(), (max(a - size, surrounding.get_first(feature)), a)
                if direction == '-' else (b, min(b + size, surrounding.get_second(feature))))

    @staticmethod
    def _create_temp_cube(cube: GenericCube, domain: DomainProperties,
                          hypercubes: Iterable[GenericCube], feature: str,
                          direction: str) -> Iterable[Expansion]:
        temp_cube, values = ITER._create_range(cube, domain, feature, direction)
        temp_cube.update_dimension(feature, values)
        overlap = temp_cube.overlap(hypercubes)
        while (overlap is not None) & (temp_cube.has_volume()):
            overlap = ITER._resolve_overlap(temp_cube, overlap, hypercubes, feature, direction)
        if (temp_cube.has_volume() & (overlap is None)) & (all(temp_cube != cube for cube in hypercubes)):
            yield Expansion(temp_cube, feature, direction)
        else:
            cube.add_limit(feature, direction)

    @staticmethod
    def _create_temp_cubes(cube: GenericCube, domain: DomainProperties,
                           hypercubes: Iterable[GenericCube]) -> Iterable[Expansion]:
        tmp_cubes = []
        for feature in domain[1].dimensions.keys():
            limit = cube.check_limits(feature)
            if limit == '*':
                continue
            for x in {'-', '+'} - {limit}:
                tmp_cubes += ITER._create_temp_cube(cube, domain, hypercubes, feature, x)
        return tmp_cubes

    def _cubes_to_update(self, dataframe: pd.DataFrame, to_expand: Iterable[GenericCube],
                         hypercubes: Iterable[GenericCube], domain: DomainProperties) \
            -> Iterable[tuple[GenericCube, Expansion]]:
        results = [(hypercube, self._best_cube(dataframe, hypercube, self._create_temp_cubes(
            hypercube, domain, hypercubes))) for hypercube in to_expand]
        return sorted([result for result in results if result[1] is not None], key=lambda x: x[1].distance)

    def _expand_or_create(self, cube: GenericCube, expansion: Expansion, hypercubes: Iterable[GenericCube]) -> None:
        if expansion.distance > self.threshold:
            hypercubes += [expansion.cube]
        else:
            cube.expand(expansion, hypercubes)

    @staticmethod
    def _find_closer_sample(dataframe: pd.DataFrame, output: float | str) -> dict[str, tuple]:
        if isinstance(output, str):
            close_sample = dataframe[dataframe.iloc[:, -1] == output].iloc[0].to_dict()
        else:
            difference = abs(dataframe.iloc[:, -1] - output)
            close_sample = dataframe[difference == min(difference)].iloc[0].to_dict()
        return close_sample

    def _generate_starting_points(self, dataframe: pd.DataFrame) -> Iterable[GenericCube]:
        if self.n_points <= 0:
            raise (Exception('InvalidAttributeValueException'))
        points: Iterable[float]
        if isinstance(dataframe.iloc[0, -1], str):
            classes = np.unique(dataframe.iloc[:, -1].values)
            points = [classes[i] for i in range(min(self.n_points, len(classes)))]
        else:
            desc = dataframe.iloc[:, -1].describe()
            min_output, max_output = desc["min"], desc["max"]
            points = [(max_output - min_output) / 2] if self.n_points == 1 else \
                [min_output + (max_output - min_output) / (self.n_points - 1) * index for index in range(self.n_points)]
        return [HyperCube.cube_from_point(ITER._find_closer_sample(dataframe, point), output=self._output)
                for point in points]

    def _initialize(self, dataframe: pd.DataFrame) -> tuple[Iterable[GenericCube], DomainProperties]:
        self._fake_dataframe = dataframe.copy()
        surrounding = HyperCube.create_surrounding_cube(dataframe, output=self._output)
        min_updates = self._calculate_min_updates(surrounding)
        self._hypercubes = self._init_hypercubes(dataframe, min_updates, surrounding)
        for hypercube in self._hypercubes:
            hypercube.update(dataframe, self.predictor)
        return self._hypercubes, (min_updates, surrounding)

    def _init_hypercubes(
            self,
            dataframe: pd.DataFrame,
            min_updates: Iterable[MinUpdate],
            surrounding: GenericCube
    ) -> Iterable[GenericCube]:
        while True:
            hypercubes = self._generate_starting_points(dataframe)
            for hypercube in hypercubes:
                hypercube.expand_all(min_updates, surrounding)
            self.n_points = self.n_points - 1
            if not HyperCube.check_overlap(hypercubes, hypercubes):
                break
        return hypercubes

    def _iterate(self, dataframe: pd.DataFrame, hypercubes: Iterable[GenericCube], domain: DomainProperties,
                 left_iteration: int) -> int:
        iterations = 0
        to_expand = [cube for cube in hypercubes if cube.limit_count < (len(dataframe.columns) - 1) * 2]
        while (len(to_expand) > 0) and (iterations < left_iteration):
            updates = list(self._cubes_to_update(dataframe, to_expand, hypercubes, domain))
            if len(updates) > 0:
                self._expand_or_create(updates[0][0], updates[0][1], hypercubes)
            iterations += 1
            to_expand = [cube for cube in hypercubes if cube.limit_count < (len(dataframe.columns) - 1) * 2]
        return iterations

    @staticmethod
    def _resolve_overlap(cube: GenericCube, overlapping_cube: GenericCube, hypercubes: Iterable[GenericCube],
                         feature: str, direction: str) -> GenericCube:
        a, b = cube[feature]
        cube.update_dimension(feature, max(overlapping_cube.get_second(feature), a) if direction == '-' else a,
                              min(overlapping_cube.get_first(feature), b) if direction == '+' else b)
        return cube.overlap(hypercubes)

    def _extract(self, dataframe: pd.DataFrame, mapping: dict[str: int] = None, sort: bool = True) -> Theory:
        self._hypercubes, domain = self._initialize(dataframe)
        temp_train = dataframe.copy()
        fake = dataframe.copy()
        iterations = 0
        while temp_train.shape[0] > 0:
            iterations += self._iterate(fake, self._hypercubes, domain, self.max_iterations - iterations)
            if (iterations >= self.max_iterations) or (not self.fill_gaps):
                break
            temp_train = temp_train.iloc[[p is None for p in self.predict(temp_train.iloc[:, :-1])]]
            if temp_train.shape[0] > 0:
                point, ratio, overlap, new_cube = temp_train.iloc[0].to_dict(), 1.0, True, None
                temp_train = temp_train.drop([temp_train.index[0]])
                while overlap is not None:
                    if new_cube is not None:
                        if not new_cube.has_volume():
                            break
                    new_cube = HyperCube.cube_from_point(point, self._output)
                    new_cube.expand_all(domain[0], domain[1], ratio)
                    overlap = new_cube.overlap(self._hypercubes)
                    ratio *= 2
                if new_cube.has_volume():
                    self._hypercubes += [new_cube]
        return self._create_theory(dataframe, sort)



from __future__ import annotations

from collections import Counter
from typing import Iterable

import numpy as np
import pandas as pd
from sklearn.cluster import DBSCAN
from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor

from psyke.clustering import InterpretableClustering
from psyke.extraction.hypercubic import Node, ClosedCube, HyperCube
from psyke.clustering.utils import select_gaussian_mixture, select_dbscan_epsilon
from psyke.utils import Target


class ExACT(InterpretableClustering):
    """
    Explanator implementing ExACT algorithm.
    """

    def __init__(self, depth: int, error_threshold: float, output: Target = Target.CONSTANT, gauss_components: int = 5):
        super().__init__(depth, error_threshold, output, gauss_components)
        self._predictor = KNeighborsClassifier() if output == Target.CLASSIFICATION else KNeighborsRegressor()
        self._predictor.n_neighbors = 1

    def __eligible_cubes(self, gauss_pred: np.ndarray, node: Node, clusters: int):
        cubes = []
        for i in range(len(np.unique(gauss_pred))):
            df = node.dataframe.iloc[np.where(gauss_pred == i)]
            if len(df) == 0:
                continue
            cubes.append(self._create_cube(df, clusters))
        indices = [self._indices(cube, node.dataframe) for cube in cubes]
        return cubes, indices

    @staticmethod
    def _indices(cube: ClosedCube, data: pd.DataFrame) -> np.ndarray | None:
        indices = cube.filter_indices(data.iloc[:, :-1])
        if len(data.iloc[indices]) * len(data.iloc[~indices]) == 0:
            return None
        return indices

    def _create_cube(self, dataframe: pd.DataFrame, clusters: int) -> ClosedCube:
        data = ExACT._remove_string_label(dataframe)
        dbscan_pred = DBSCAN(eps=select_dbscan_epsilon(data, clusters)).fit_predict(data.iloc[:, :-1])
        return HyperCube.create_surrounding_cube(
            dataframe.iloc[np.where(dbscan_pred == Counter(dbscan_pred).most_common(1)[0][0])],
            True, self._output
        )

    def extract(self, dataframe: pd.DataFrame) -> Iterable[HyperCube]:
        self._predictor.fit(dataframe.iloc[:, :-1], dataframe.iloc[:, -1])
        self._hypercubes = \
            self._iterate(Node(dataframe, HyperCube.create_surrounding_cube(dataframe, True, self._output)))
        return list(self._hypercubes)

    def print(self):
        for cube in self._hypercubes:
            print(f'Output is {cube.output} if:')
            for feature in cube.dimensions:
                lower, upper = cube[feature]
                print(f'    {feature} is in [{lower:.2f}, {upper:.2f}]')

    @staticmethod
    def _remove_string_label(dataframe: pd.DataFrame):
        return dataframe.replace({dataframe.columns[-1]: {v: k for k, v in dict(
            enumerate(dataframe.iloc[:, -1].unique())
        ).items()}}) if isinstance(dataframe.iloc[0, -1], str) else dataframe

    def _iterate(self, surrounding: Node) -> Iterable[HyperCube]:
        to_split = [(self.error_threshold * 10, 1, 1, surrounding)]
        while len(to_split) > 0:
            to_split.sort(reverse=True)
            (_, depth, _, node) = to_split.pop()
            data = ExACT._remove_string_label(node.dataframe)
            gauss_params = select_gaussian_mixture(data, self.gauss_components)
            gauss_pred = gauss_params[2].predict(data)
            cubes, indices = self.__eligible_cubes(gauss_pred, node, gauss_params[1])
            cubes = [(c.volume(), len(idx), i, idx, c) for i, (c, idx) in enumerate(zip(cubes, indices))
                     if (idx is not None) and (not node.cube.equal(c))]
            if len(cubes) < 1:
                continue
            _, _, _, indices, cube = max(cubes)

            cube.update(node.dataframe[indices], self._predictor)
            node.right = Node(node.dataframe[indices], cube)
            node.cube.update(node.dataframe[~indices], self._predictor)
            node.left = Node(node.dataframe[~indices], node.cube)

            if depth < self.depth and cube.diversity > self.error_threshold:
                to_split.append((cube.diversity, depth + 1, np.random.uniform(), node.right))
        return self._node_to_cubes(surrounding)

    def _node_to_cubes(self, root: Node) -> list[ClosedCube]:
        if root.right is None:
            return [root.cube]
        else:
            return self._node_to_cubes(root.right) + self._node_to_cubes(root.left)

    @property
    def n_rules(self):
        return len(list(self._hypercubes))



import numpy as np
import pandas as pd
from kneed import KneeLocator
from sklearn.cluster import DBSCAN
from sklearn.mixture import GaussianMixture
from sklearn.neighbors import NearestNeighbors


def select_gaussian_mixture(data: pd.DataFrame, max_components) -> tuple[float, int, GaussianMixture]:
    components = range(2, max_components + 1)
    try:
        models = [GaussianMixture(n_components=n).fit(data) for n in components if n <= len(data)]
    except ValueError:
        print(data)
        print(len(data))
    return min([(m.bic(data) / (i + 2), (i + 2), m) for i, m in enumerate(models)])


def select_dbscan_epsilon(data: pd.DataFrame, clusters: int) -> float:
    neighbors = NearestNeighbors(n_neighbors=min(len(data.columns) * 2, len(data))).fit(data)
    distances = sorted(np.mean(neighbors.kneighbors(data)[1], axis=1), reverse=True)
    try:
        kn = KneeLocator([d for d in range(len(distances))], distances,
                         curve='convex', direction='decreasing', online=True)
        if kn.knee_y is None:
            epsilon = max(distances[-1], 1e-3)
        else:
            epsilon = kn.knee_y
    except (RuntimeWarning, UserWarning, ValueError):
        epsilon = max(distances[-1], 1e-3)
    k = 1.
    dbscan_pred = DBSCAN(eps=epsilon * k).fit_predict(data.iloc[:, :-1])
    # while Counter(dbscan_pred).most_common(1)[0][0] == -1:
    for i in range(1000):
        if len(np.unique(dbscan_pred)) < clusters + 1:
            break
        k += .1
        dbscan_pred = DBSCAN(eps=epsilon * k).fit_predict(data.iloc[:, :-1])
    return epsilon * k



from typing import Callable, Any


class SortedList(list):

    def __init__(self, comparator: Callable[[Any, Any], int]):
        super().__init__()
        self.comparator = comparator

    def add(self, item) -> None:
        if len(self) == 0:
            self.insert(0, item)
        else:
            starting_len = len(self)
            for index, element in enumerate(self):
                if self.comparator(element, item) > 0:
                    self.insert(index, item)
                    break
            if len(self) == starting_len:
                self.append(item)

    def add_all(self, other) -> None:
        for item in other:
            self.add(item)



import random

import pandas as pd
from matplotlib import pyplot as plt
from sklearn.datasets import load_iris, load_wine, fetch_california_housing
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.linear_model import LinearRegression
from sklearn.model_selection import train_test_split
from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor
from sklearn.svm import SVC, SVR
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor

from psyke import Extractor
from psyke.extraction.hypercubic import Grid, FixedStrategy, FeatureRanker
from psyke.extraction.hypercubic.strategy import AdaptiveStrategy
from psyke.gui.model import PREDICTORS, FIXED_PREDICTOR_PARAMS, EXTRACTORS, cast_param, DatasetError, PredictorError
from psyke.gui.view import COLOR_MAPS
from psyke.gui.view.plot import init_plot, plotSamples, create_grid, plot_regions
from psyke.utils import Target
from psyke.utils.dataframe import get_discrete_features_supervised, get_discrete_dataset, get_scaled_dataset, \
    scale_dataset


class Model:

    def __init__(self, path):
        self.path = f'{path}/plots'
        self.task = 'Classification'
        self.preprocessing_action = None
        self.preprocessing = None
        self.dataset = None
        self.data = None
        self.pruned_data = None
        self.ranked_data = None
        self.colormap = None
        self.train = None
        self.test = None
        self.predictor_name = None
        self.predictor = None
        self.predictor_params = {}
        self.extractor_name = None
        self.extractor = None
        self.extractor_params = {}
        self.theory = None
        self.data_plot = None
        self.predictor_plot = None
        self.extractor_plot = None

    def select_task(self, task):
        self.task = task
        self.colormap = None

    def select_preprocessing(self, action):
        self.preprocessing_action = action

    def select_colormap(self, cmap):
        self.colormap = cmap

    def select_dataset(self, dataset):
        self.dataset = dataset

    def select_predictor(self, predictor):
        self.predictor_name = predictor

    def select_extractor(self, extractor):
        self.extractor_name = extractor

    def reset_preprocessing(self):
        self.preprocessing_action = None
        self.preprocessing = None

    def reset_dataset(self, soft=False):
        if not soft:
            self.data = None
        self.pruned_data = None
        self.train = None
        self.test = None
        self.ranked_data = None
        self.data_plot = None

    def reset_predictor(self):
        self.predictor_name = None
        self.predictor = None
        self.predictor_params = {}
        self.predictor_plot = None

    def reset_extractor(self):
        self.extractor_name = None
        self.extractor = None
        self.extractor_params = {}
        self.theory = None
        self.extractor_plot = None

    def load_dataset(self, ret=False):
        print(f'Loading {self.dataset}... ', end='')
        if self.dataset == 'Arti':
            data = pd.read_csv('test/resources/datasets/arti.csv')
        elif self.dataset == 'Bank marketing':
            data = pd.read_csv('test/resources/datasets/akbilgic.csv')
            data = data.iloc[:, [i for i in range(len(data.columns) - 7, len(data.columns))] + [1]]
        else:
            if self.dataset == 'Iris':
                x, y = load_iris(return_X_y=True, as_frame=True)
                x.columns = ['SepalLength', 'SepalWidth', 'PetalLength', 'PetalWidth']
                y.name = 'iris'
                data = (x, y.replace({0: 'setosa', 1: 'versicolor', 2: 'virginica'}))
            elif self.dataset == 'Wine':
                x, y = load_wine(return_X_y=True, as_frame=True)
                data = (x, y.apply(str))
            elif self.dataset == "House":
                data = fetch_california_housing(return_X_y=True, as_frame=True)
            else:
                raise DatasetError
            data = data[0].join(data[1])
        print('Done')
        if ret:
            return data
        self.ranked_data = FeatureRanker(data.columns[:-1]).fit_on_data(data).rankings()
        if self.preprocessing_action == 'Discretize':
            self.preprocessing = get_discrete_features_supervised(data)
            self.data = get_discrete_dataset(data.iloc[:, :-1], self.preprocessing, False).join(data.iloc[:, -1])
        elif self.preprocessing_action == 'Scale':
            self.data, self.preprocessing = get_scaled_dataset(data)
        else:
            self.data = data

    def select_features(self, features):
        inputs = [k for k, v in features.items() if v == 'I']
        output = [k for k, v in features.items() if v == 'O'][0]
        if self.preprocessing_action == 'Discretize':
            inputs = [[list(discretization.admissible_values.keys()) for discretization in self.preprocessing
                       if discretization.name == variable] for variable in inputs]
            inputs = [item for sublist in inputs for item in sublist[0]]
        self.pruned_data = self.data[inputs].join(self.data[output])
        data = self.load_dataset(True)
        try:
            self.ranked_data = FeatureRanker(data[inputs]).fit_on_data(data[inputs].join(data[output])).rankings()
        except ValueError:
            self.ranked_data = None
            raise ValueError

    def train_predictor(self):
        self.read_predictor_param()

        print(f'Training {self.predictor_name}... ', end='')
        if self.predictor_name == 'K-NN':
            self.predictor = KNeighborsClassifier() if self.task == 'Classification' else KNeighborsRegressor()
            self.predictor.n_neighbors = self.predictor_params['K']
        elif self.predictor_name == 'LR':
            self.predictor = LinearRegression()
        elif self.predictor_name in ['DT', 'RF']:
            if self.predictor_name == 'DT':
                self.predictor = DecisionTreeClassifier() if self.task == 'Classification' else DecisionTreeRegressor()
            else:
                self.predictor = RandomForestClassifier() if self.task == 'Classification' else RandomForestRegressor()
                self.predictor.n_estimators = self.predictor_params['N estimators']
            self.predictor.max_depth = self.predictor_params['Max depth']
            self.predictor.max_leaf_nodes = self.predictor_params['Max leaves']
        elif self.predictor_name == 'SVM':
            self.predictor = SVC() if self.task == 'Classification' else SVR(epsilon=self.predictor_params['Epsilon'])
            self.predictor.C = self.predictor_params['Regularization']
            self.predictor.kernel = self.predictor_params['Kernel'].lower()
        else:
            raise PredictorError
        self.train, self.test = train_test_split(self.data if self.pruned_data is None else self.pruned_data,
                                                 test_size=self.predictor_params['Test set'],
                                                 random_state=self.predictor_params['Split seed'])
        self.predictor.fit(self.train.iloc[:, :-1], self.train.iloc[:, -1])
        print('Done')

    def train_extractor(self):
        def get_output():
            return Target.CONSTANT if self.extractor_params['Constant output'] else Target.REGRESSION

        def get_rankings():
            data = (self.data if self.pruned_data is None else self.pruned_data)
            return FeatureRanker(data.columns[:-1]).fit_on_data(data).rankings()

        def get_strategy():
            return FixedStrategy(self.extractor_params['Splits']) if not self.extractor_params['Adaptive'] else \
                AdaptiveStrategy(get_rankings(), [(self.extractor_params['Adaptive threshold'],
                                                   self.extractor_params['Splits'])])

        # GRIDEX, GRIDREX -> strategy
        self.read_extractor_param()

        print(f'Training {self.extractor_name}... ', end='')
        if self.extractor_name == 'REAL':
            self.extractor = Extractor.real(self.predictor, discretization=self.preprocessing)
        elif self.extractor_name == 'Trepan':
            self.extractor = Extractor.trepan(self.predictor, min_examples=self.extractor_params['Min examples'],
                                              max_depth=self.extractor_params['Max depth'],
                                              discretization=self.preprocessing)
        elif self.extractor_name == 'CART':
            discretization = self.preprocessing if self.preprocessing_action == 'Discretize' else None
            normalization = self.preprocessing if self.preprocessing_action == 'Scale' else None
            self.extractor = Extractor.cart(self.predictor, max_depth=self.extractor_params['Max depth'],
                                            max_leaves=self.extractor_params['Max leaves'],
                                            simplify=self.extractor_params['Simplify'],
                                            discretization=discretization, normalization=normalization)
        elif self.extractor_name == 'Iter':
            self.extractor = Extractor.iter(self.predictor, threshold=self.extractor_params['Threshold'],
                                            min_examples=self.extractor_params['Min examples'],
                                            min_update=self.extractor_params['Min update'],
                                            n_points=self.extractor_params['N points'],
                                            max_iterations=self.extractor_params['Max iterations'],
                                            fill_gaps=self.extractor_params['Fill gaps'],
                                            normalization=self.preprocessing)
        elif self.extractor_name == 'GridEx':
            self.extractor = Extractor.gridex(self.predictor, threshold=self.extractor_params['Threshold'],
                                              min_examples=self.extractor_params['Min examples'],
                                              grid=Grid(self.extractor_params['Max depth'], get_strategy()),
                                              normalization=self.preprocessing)
        elif self.extractor_name == 'GridREx':
            self.extractor = Extractor.gridrex(self.predictor, threshold=self.extractor_params['Threshold'],
                                               min_examples=self.extractor_params['Min examples'],
                                               grid=Grid(self.extractor_params['Max depth'], get_strategy()),
                                               normalization=self.preprocessing)
        elif self.extractor_name in ['CReEPy', 'ORCHiD']:
            extractor = Extractor.creepy if self.extractor_name == 'CReEPy' else Extractor.orchid
            self.extractor = extractor(self.predictor, depth=self.extractor_params['Max depth'],
                                       error_threshold=self.extractor_params['Threshold'],
                                       ignore_threshold=self.extractor_params['Feat threshold'],
                                       gauss_components=self.extractor_params['Max components'],
                                       output=Target.CLASSIFICATION if self.task == 'Classification' else get_output(),
                                       ranks=get_rankings(), normalization=self.preprocessing)
        else:
            raise NotImplementedError

        self.theory = self.extractor.extract(self.train)
        print('Done')

    def plot(self, inputs, output, save=False):
        x = inputs[0]
        y = inputs[1] if len(inputs) > 1 else output
        z = output if len(inputs) > 1 else None

        index = 0 if self.colormap is None else \
            [i for i, (n, _) in enumerate(COLOR_MAPS[self.task]) if n == self.colormap][0]
        cmap = COLOR_MAPS[self.task][index][1]

        data = self.load_dataset(True)
        actual_data = self.data if self.pruned_data is None else self.pruned_data

        init_plot(data[x], data[y], 'Data set')
        plotSamples(data[x], data[y], data[z if z is not None else y], cmap, index,
                    f'{self.path}/data.pdf' if save else None)
        self.data_plot = plt.gcf()
        plt.close()

        if self.predictor is None:
            return

        grid = create_grid(x, y, data)
        processed_grid = get_discrete_dataset(grid, self.preprocessing, False) if \
            self.preprocessing_action == 'Discretize' else scale_dataset(grid, self.preprocessing) \
            if self.preprocessing_action == 'Scale' else grid
        processed_grid = processed_grid[actual_data.columns[:-1]]

        for model, name in zip([self.predictor, self.extractor], [self.predictor_name, self.extractor_name]):
            if model is None:
                break
            init_plot(data[x], data[y], name)
            predictions = model.predict(processed_grid)
            if self.preprocessing_action == 'Scale':
                m, s = self.preprocessing[actual_data.columns[-1]]
                predictions = predictions * s + m
            grid_data = pd.concat(
                [grid, pd.DataFrame(predictions, columns=[actual_data.columns[-1]])], axis=1
            )
            grid_data = grid_data[grid_data.iloc[:, -1].notna()]
            grouped = grid_data.groupby([x, y])[grid_data.columns[-1]]
            outputs = grouped.agg(pd.Series.mode) if isinstance(grid_data.iloc[0, -1], str) else grouped.mean()
            grid_data = grid_data.groupby([x, y])[grid_data.columns[:-1]].mean().join(outputs)
            if z is not None:
                plot_regions(grid_data[x].values, grid_data[y].values, grid_data[z].values, cmap)
            plotSamples(data[x], data[y], data[z if z is not None else y], cmap, index,
                        f'{self.path}/{name}.pdf' if save else None)
            if isinstance(model, Extractor):
                self.extractor_plot = plt.gcf()
            else:
                self.predictor_plot = plt.gcf()
            plt.close()

    def set_predictor_param(self, key, value):
        self.predictor_params[key] = \
            cast_param(dict(FIXED_PREDICTOR_PARAMS, **PREDICTORS[self.predictor_name][1]), key, value)

    def set_extractor_param(self, key, value):
        self.extractor_params[key] = cast_param(EXTRACTORS[self.extractor_name][1], key, value)

    def read_predictor_param(self):
        all_params = dict(FIXED_PREDICTOR_PARAMS, **PREDICTORS[self.predictor_name][1])
        for name, (default, _, constraint) in all_params.items():
            if constraint is not None and constraint != self.task:
                continue
            if name not in self.predictor_params.keys():
                self.predictor_params[name] = default

    def read_extractor_param(self):
        params = EXTRACTORS[self.extractor_name][1]
        for name, (default, _, constraint) in params.items():
            if constraint is not None and constraint != self.task:
                continue
            if name not in self.extractor_params.keys():
                self.extractor_params[name] = default



from pathlib import Path
from typing import TextIO

PATH = Path(__file__).parents[0]


def get_dataset_path(filename: str) -> Path:
    return PATH / f"{filename}.csv"


def open_dataset(filename: str) -> TextIO:
    return open(get_dataset_path(filename))




from array import array
from typing import Callable, Iterable
import numpy as np
import pandas as pd
from matplotlib import colors
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from tuprolog.solve.prolog import prolog_solver
from tuprolog.theory import Theory, mutable_theory
from psyke.utils.logic import data_to_struct, pretty_theory, get_in_rule, get_not_in_rule

import matplotlib
matplotlib.use('TkAgg')


def predict_from_theory(theory: Theory, data: pd.DataFrame) -> list[float or str]:
    solver = prolog_solver(static_kb=mutable_theory(theory).assertZ(get_in_rule()).assertZ(get_not_in_rule()))
    index = data.shape[1] - 1
    y_element = data.iloc[0, -1]
    cast: Callable = lambda x: (str(x) if isinstance(y_element, str) else x)
    substitutions = [solver.solveOnce(data_to_struct(data)) for _, data in data.iterrows()]
    return [cast(query.solved_query.get_arg_at(index)) if query.is_yes else -1 for query in substitutions]


def plot_theory(theory: Theory, data: pd.DataFrame = None, output: str = 'plot.pdf', azimuth: float = 45,
                distance: float = 9, elevation: float = 5, show_theory: bool = True, features: Iterable[str] = None) -> None:
    # Check if the number of common variables in clauses is less or equal to three.
    # If not raise an exception.
    fresh_theory = mutable_theory(theory)
    clauses = fresh_theory.clauses
    variables = sorted(list(set(arg.args[0].name.split('_')[0] for clause in clauses if clause.body_size > 0 and clause.body.is_recursive for arg in clause.body.unfolded)), reverse=True)
    if len(variables) > 3:
        raise Exception("Theory contains too many different features in the body of clauses, maximum is 3.")
    # If data is None, then create synthetic data covering a good portion of the variables space.
    # Just skip for now.
    if data is None:
        raise Exception("Method without data is not implemented yet")

    # Prepare data
    ys = predict_from_theory(fresh_theory, data)
    xs = data[variables].values.tolist()
    for i in range(len(ys)):
        xs[i].append(ys[i])

    # Prepare colors
    if isinstance(ys[0], str):
        np.warnings.filterwarnings('ignore', category=np.VisibleDeprecationWarning)

        class ColorGenerator:

            def __init__(self):
                self.color_list = ['red', 'royalblue', 'green', 'orange', 'pink', 'acqua', 'grey']
                self.counter = 0

            def get_new_color(self) -> str:
                self.counter += 1
                if self.counter > len(self.color_list):
                    raise Exception("Classes exceed the maximum supported number (7)")
                return self.color_list[self.counter - 1]

        classes = set(ys)
        generator = ColorGenerator()
        class_color = {c: generator.get_new_color() for c in classes}
        get_color: Callable = lambda c: class_color[c]
    else:
        def color_fader(v: float = 0., c1: str = 'green', c2: str = 'red'):
            c1 = array(colors.to_rgb(c1))
            c2 = array(colors.to_rgb(c2))
            return colors.to_hex((1 - v) * c1 + v * c2)
        min_value = min(ys)
        max_value = max(ys)
        get_normalized_value: Callable = lambda v: (v - min_value)/(max_value - min_value)
        get_color: Callable = lambda c: color_fader(get_normalized_value(c))

    fig = plt.figure()
    fig.set_size_inches(10, 10)
    if len(variables) == 3:
        ax = fig.add_subplot(projection='3d')
    else:
        ax = fig.add_subplot()

    for x in xs:
        ax.scatter(*x[:-1], c=get_color(x[-1]), s=14)

    ax.set_xlabel(variables[0], fontsize=18)
    ax.set_ylabel(variables[1], fontsize=18)
    if len(variables) == 3:
        ax.set_zlabel(variables[2], fontsize=18)

    ax.azim = azimuth
    ax.dist = distance
    ax.elev = elevation
    ax.set_title('Predictions according to Prolog theory', fontsize=24)
    if show_theory:
        pass
        # ax.text2D(0., 0.88, pretty_theory(theory, new_line=False), transform=ax.transAxes, fontsize=8)
    if isinstance(ys[0], str):
        custom_lines = [Line2D([0], [0], marker='o', markerfacecolor=get_color(c), markersize=20, color='w') for c in classes]
        ax.legend(custom_lines, classes, loc='upper left', numpoints=1, ncol=3, fontsize=18, bbox_to_anchor=(0, 0))
    plt.savefig(output, format='pdf')



from __future__ import annotations

from functools import reduce
from typing import Iterable


class Strategy:
    def __init__(self):
        self._partitions = None

    def get(self, feature: str) -> int:
        raise NotImplementedError

    def partition_number(self, features: Iterable[str]) -> int:
        return reduce(lambda x, y: x * y, map(self.get, features), 1)

    def equals(self, strategy, features: Iterable[str]) -> bool:
        eq = True
        for f in features:
            eq = eq and self.get(f) == strategy.get(f)
        return eq

    def __str__(self):
        return self._partitions

    def __repr__(self):
        return self.__str__()


class FixedStrategy(Strategy):
    def __init__(self, partitions: int = 2):
        super().__init__()
        self._partitions = partitions

    def get(self, feature: str) -> int:
        return self._partitions

    def __str__(self):
        return "Fixed ({})".format(super().__str__())


class AdaptiveStrategy(Strategy):
    def __init__(self, features: Iterable[str], partitions: Iterable[tuple[float, float]] | None = None):
        super().__init__()
        self.features = features
        self._partitions = partitions if partitions is not None else [(0.33, 2), (0.67, 3)]

    def get(self, feature: str) -> int:
        importance = next(filter(lambda t: t[0] == feature, self.features))[1]
        n = 1
        for (imp, part) in self._partitions:
            if importance >= imp:
                n = part
            else:
                break
        return n

    def __str__(self):
        return "Adaptive ({})".format(super().__str__())



import numpy as np
from psyke.gui.view.layout import PanelBoxLayout, CoupledRelativeLayout, CoupledToggleButton
from psyke.gui.model import TASKS, DATASETS
from psyke.gui.view import DATASET_MESSAGE, INFO_DATASET_MESSAGE, INFO_DATASET_PREFIX


class DataPanel(PanelBoxLayout):

    def __init__(self, controller, ratio=1, **kwargs):
        super().__init__(controller, 'Load', INFO_DATASET_MESSAGE, 2, ratio, **kwargs)

        task_panel = CoupledRelativeLayout(1, ratio)
        for i, task in enumerate(TASKS):
            btn_task = CoupledToggleButton(i, text=task, group='task', state='down'
                                           if self.controller.get_task_from_model() == task else 'normal')
            btn_task.bind(state=self.select_task)
            task_panel.add_widget(btn_task)

        processing_panel = CoupledRelativeLayout(3, ratio)
        self.discretize_button = CoupledToggleButton(0, text='Discretize')
        self.scale_button = CoupledToggleButton(1, text='Scale')
        for btn_proc in [self.discretize_button, self.scale_button]:
            btn_proc.state = 'normal'
            btn_proc.group = 'preprocessing'
            btn_proc.bind(state=self.select_preprocessing)
            processing_panel.add_widget(btn_proc)

        self.add_widget(task_panel)
        self.add_widget(self.main_panel)
        self.add_widget(processing_panel)
        self.add_widget(self.info_label)

    def select(self, spinner, text):
        self.controller.select_dataset(text if text != DATASET_MESSAGE else None)
        self.go_button.disabled = False

    def go_action(self, button):
        self.controller.load_dataset()

    def enable(self):
        self.discretize_button.disabled = False
        self.scale_button.disabled = False

    def disable(self):
        self.discretize_button.disabled = True
        self.scale_button.disabled = True

    def set_info(self):
        data, pruned_data, _, _ = self.controller.get_data_from_model()
        if pruned_data is not None:
            data = pruned_data
        self.info_label.text = INFO_DATASET_MESSAGE if data is None else \
            INFO_DATASET_PREFIX + \
            f'Dataset: {self.controller.get_dataset_from_model()}\n' \
            f'Input variables: {len(data.columns) - 1}\nInstances: {len(data)}'
        if data is not None and isinstance(data.iloc[0, -1], str):
            self.info_label.text += f'\nClasses: {len(np.unique(data.iloc[:, -1]))}'

    def init(self):
        self.spinner_options.text = DATASET_MESSAGE
        self.spinner_options.values = [entry[0] for entry in DATASETS
                                       if self.controller.get_task_from_model() in entry[1]]
        self.go_button.disabled = True
        self.discretize_button.state = 'normal'
        self.scale_button.state = 'normal'
        self.disable()

    def select_task(self, button, value):
        if value == 'down':
            self.controller.select_task(button.text)

    def select_preprocessing(self, button, value):
        self.controller.select_preprocessing(button.text, value == 'down')



from __future__ import annotations
from abc import ABC
from typing import Iterable
import numpy as np
import pandas as pd
from sklearn.base import ClassifierMixin, RegressorMixin
from sklearn.feature_selection import SelectKBest, f_regression, f_classif
from sklearn.linear_model import LinearRegression
from tuprolog.core import Var, Struct, clause
from tuprolog.theory import Theory, mutable_theory
from psyke import Extractor, logger
from psyke.extraction.hypercubic.hypercube import HyperCube, RegressionCube, ClassificationCube, ClosedCube
from psyke.utils.logic import create_variable_list, create_head, to_var, Simplifier
from psyke.utils import Target, get_int_precision
from psyke.extraction.hypercubic.strategy import Strategy, FixedStrategy


class HyperCubeExtractor(Extractor, ABC):

    def __init__(self, predictor, normalization):
        super().__init__(predictor, normalization=normalization)
        self._hypercubes = []
        self._output = Target.CONSTANT

    def _predict(self, dataframe: pd.DataFrame) -> Iterable:
        return np.array([self._predict_from_cubes(dict(row.to_dict())) for _, row in dataframe.iterrows()])

    def _predict_from_cubes(self, data: dict[str, float]) -> float | None:
        data = {k: v for k, v in data.items()}
        for cube in self._hypercubes:
            if cube.__contains__(data):
                if self._output == Target.CLASSIFICATION:
                    return HyperCubeExtractor._get_cube_output(cube, data)
                else:
                    return round(HyperCubeExtractor._get_cube_output(cube, data), get_int_precision())
        return None

    def _default_cube(self) -> HyperCube | RegressionCube | ClassificationCube:
        if self._output == Target.CONSTANT:
            return HyperCube()
        if self._output == Target.REGRESSION:
            return RegressionCube()
        return ClassificationCube()

    @staticmethod
    def _get_cube_output(cube: HyperCube | RegressionCube, data: dict[str, float]) -> float:
        return cube.output.predict(pd.DataFrame([data])).flatten()[0] if \
            isinstance(cube, RegressionCube) else cube.output

    @staticmethod
    def _create_head(dataframe: pd.DataFrame, variables: list[Var], output: float | LinearRegression) -> Struct:
        return create_head(dataframe.columns[-1], variables[:-1], output) \
            if not isinstance(output, LinearRegression) else \
            create_head(dataframe.columns[-1], variables[:-1], variables[-1])

    def _ignore_dimensions(self) -> Iterable[str]:
        return []

    def _create_theory(self, dataframe: pd.DataFrame, sort: bool = True) -> Theory:
        new_theory = mutable_theory()
        for cube in self._hypercubes:
            logger.info(cube.output)
            logger.info(cube.dimensions)
            variables = create_variable_list([], dataframe, sort)
            variables[dataframe.columns[-1]] = to_var(dataframe.columns[-1])
            head = HyperCubeExtractor._create_head(dataframe, list(variables.values()),
                                                   self.unscale(cube.output, dataframe.columns[-1]))
            body = cube.body(variables, self._ignore_dimensions(), self.unscale, self.normalization)
            new_theory.assertZ(clause(head, body))
        return HyperCubeExtractor._prettify_theory(new_theory)

    @staticmethod
    def _prettify_theory(theory: Theory) -> Theory:
        visitor = Simplifier()
        new_clauses = []
        for c in theory.clauses:
            body = c.body
            structs = body.unfolded if c.body_size > 1 else [body]
            new_structs = []
            for s in structs:
                new_structs.append(s.accept(visitor))
            new_clauses.append(clause(c.head, new_structs))
        return mutable_theory(new_clauses)

    @property
    def n_rules(self):
        return len(self._hypercubes)


class FeatureRanker:
    def __init__(self, feat):
        self.scores = None
        self.feat = feat

    def fit(self, model, samples):
        predictions = np.array(model.predict(samples)).flatten()
        function = f_classif if isinstance(model, ClassifierMixin) else f_regression
        best = SelectKBest(score_func=function, k="all").fit(samples, predictions)
        self.scores = np.array(best.scores_) / max(best.scores_)
        return self

    def fit_on_data(self, samples):
        function = f_classif if isinstance(samples.iloc[0, -1], str) else f_regression
        best = SelectKBest(score_func=function, k="all").fit(samples.iloc[:, :-1], samples.iloc[:, -1])
        self.scores = np.array(best.scores_) / max(best.scores_)
        return self

    def rankings(self):
        return list(zip(self.feat, self.scores))


class Grid:
    def __init__(self, iterations: int = 1, strategy: Strategy | list[Strategy] = FixedStrategy()):
        self.iterations = iterations
        self.strategy = strategy

    def get(self, feature: str, depth: int) -> int:
        if isinstance(self.strategy, list):
            return self.strategy[depth].get(feature)
        else:
            return self.strategy.get(feature)

    def iterate(self) -> range:
        return range(self.iterations)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return "Grid ({}). {}".format(self.iterations, self.strategy)


class Node:
    def __init__(self, dataframe: pd.DataFrame, cube: ClosedCube = None):
        self.dataframe = dataframe
        self.cube: ClosedCube = cube
        self.right: Node | None = None
        self.left: Node | None = None

    @property
    def children(self) -> list[Node]:
        return [self.right, self.left]

    def search(self, point: dict[str, float]) -> ClosedCube:
        if self.right is None:
            return self.cube
        if self.right.cube.__contains__(point):
            return self.right.search(point)
        return self.left.search(point)

    @property
    def leaves(self):
        if self.right is None:
            return 1
        return self.right.leaves + self.left.leaves



from __future__ import annotations
from itertools import chain
from typing import Iterable, Any
import pandas as pd


class Node:

    def __init__(self, samples: pd.DataFrame, n_examples: int, constraints: Iterable[tuple[str, float]] = None,
                 children: list[Node] = None, depth: int = 0):
        self.samples = samples
        self.n_examples = n_examples
        self.constraints = [] if constraints is None else constraints
        self.children = [] if children is None else children
        self.depth = depth

    def __str__(self):
        name = ''.join(('' if c[1] > 0 else '!') + c[0] + ', ' for c in self.constraints)
        return name[:-2] + ' = ' + str(self.dominant)

    @property
    def priority(self) -> float:
        return -(self.reach * (1 - self.fidelity))

    @property
    def fidelity(self) -> float:
        return 1.0 * self.correct / (self.samples.shape[0] if self.samples.shape[0] > 0 else 1)

    @property
    def reach(self) -> float:
        return 1.0 * self.samples.shape[0] / self.n_examples

    @property
    def correct(self) -> float:
        return sum(self.samples.iloc[:, -1] == self.dominant)

    @property
    def dominant(self) -> Any:
        return self.samples.iloc[:, -1].mode()[0] if self.samples.shape[0] > 0 else ''

    @property
    def n_classes(self) -> int:
        return len(set(self.samples.iloc[:, -1]))

    def __iter__(self) -> Iterable[Node]:
        for child in chain(*map(iter, self.children)):
            yield child
        yield self


class Split:

    # TODO: should be configurable by user
    PRIORITY_BONUS: int = 100
    PRIORITY_PENALTY: int = 200

    def __init__(self, parent: Node, children: tuple[Node, Node]):
        self.parent = parent
        self.children = children

    @property
    def priority(self) -> float:
        return self.__priority(self.parent)

    def __priority(self, parent: Node) -> float:
        true_node, false_node = self.children
        priority = - (true_node.fidelity + false_node.fidelity)
        for node in [true_node, false_node]:
            priority -= self.PRIORITY_BONUS if parent.n_classes > node.n_classes else 0
        priority += self.PRIORITY_PENALTY if true_node.dominant == false_node.dominant else 0
        return priority


class SplitLogic:

    DEFAULT = 1



from kivy.uix.label import Label
from kivy.uix.relativelayout import RelativeLayout

from psyke.gui.view import THEORY_PREFIX, THEORY_MESSAGE, THEORY_ERROR_MESSAGE
from psyke.utils.logic import pretty_theory


class TheoryPanel(RelativeLayout):

    def __init__(self, controller, **kwargs):
        super().__init__(**kwargs)
        self.controller = controller

        self.add_widget(Label(text=THEORY_PREFIX, size_hint=(1, .17), pos_hint={'x': -.42, 'y': .77}))
        self.theory_label = Label(size_hint=(1, .83))
        self.add_widget(self.theory_label)

    def init(self):
        self.theory_label.text = THEORY_MESSAGE
        self.theory_label.pos_hint = {'x': -.2, 'y': 0.}

    def set_info(self):
        theory = self.controller.get_theory_from_model()
        if theory is not None:
            n_rules = len(list(theory.clauses))
            theory = pretty_theory(theory)
            print(theory)
            if n_rules > 6:
                self.theory_label.text = THEORY_ERROR_MESSAGE['amount']
            else:
                length = len(max(theory.split('\n'), key=len))
                if length > 150:
                    self.theory_label.text = THEORY_ERROR_MESSAGE['length']
                else:
                    self.theory_label.text = theory + '\n' * (6 - n_rules) * 2
                    self.theory_label.pos_hint = {'x': 0., 'y': 0.}
        else:
            self.init()




from __future__ import annotations
from collections import Iterable
import numpy as np
import pandas as pd
from sklearn.base import ClassifierMixin
from tuprolog.core import clause
from tuprolog.theory import Theory
from psyke import Extractor, PedagogicalExtractor
from psyke.extraction.hypercubic import HyperCubeExtractor
from psyke.utils import Target


class CReEPy(PedagogicalExtractor, HyperCubeExtractor):
    """
    Explanator implementing CReEPy algorithm.
    """

    def __init__(self, predictor, depth: int, error_threshold: float, output: Target = Target.CONSTANT,
                 gauss_components: int = 5, ranks: list[(str, float)] = [], ignore_threshold: float = 0.0,
                 normalization=None, clustering=Extractor.exact):
        super().__init__(predictor, normalization)
        self._output = Target.CLASSIFICATION if isinstance(predictor, ClassifierMixin) else output
        self.clustering = clustering(depth, error_threshold, self._output, gauss_components)
        self.ranks = ranks
        self.ignore_threshold = ignore_threshold

    def _extract(self, dataframe: pd.DataFrame, mapping: dict[str: int] = None, sort: bool = True) -> Theory:
        self._hypercubes = self.clustering.extract(dataframe)
        for cube in self._hypercubes:
            for dimension in self._ignore_dimensions():
                cube[dimension] = [-np.inf, np.inf]
        theory = self._create_theory(dataframe)
        last_clause = list(theory.clauses)[-1]
        theory.retract(last_clause)
        theory.assertZ(clause(
            last_clause.head, [list(last_clause.body)[-1]] if self._output is Target.REGRESSION else []))
        last_cube = self._hypercubes[-1]
        for dimension in last_cube.dimensions.keys():
            last_cube[dimension] = [-np.inf, np.inf]
        return theory

    def _ignore_dimensions(self) -> Iterable[str]:
        return [dimension for dimension, relevance in self.ranks if relevance < self.ignore_threshold]

    @property
    def n_rules(self):
        return len(list(self._hypercubes))



import unittest
import pandas as pd

from psyke.extraction.hypercubic.hypercube import FeatureNotFoundException
from psyke.extraction.hypercubic.utils import MinUpdate, Expansion
from psyke.utils import get_int_precision
from sklearn.neighbors import KNeighborsRegressor
from test.psyke import Predictor
from psyke.extraction.hypercubic import HyperCube
from test.resources.datasets import get_dataset_path


class AbstractTestHypercube(unittest.TestCase):

    def setUp(self):
        self.dimensions = {'X': (0.2, 0.6), 'Y': (0.7, 0.9)}
        self.mean = 0.5
        self.cube = HyperCube(self.dimensions, set(), self.mean)
        cubes = [({'X': (6.4, 7.9), 'Y': (5.7, 8.9)}, 5.3),
                 ({'X': (0.7, 0.8), 'Y': (0.75, 0.85)}, 6.1),
                 ({'X': (6.6, 7.0), 'Y': (9.1, 10.5)}, 7.5)]
        self.hypercubes = [HyperCube(cube[0], set(), cube[1]) for cube in cubes]
        self.dataset = pd.read_csv(get_dataset_path('arti'))
        self.filtered_dataset = self.dataset[self.dataset.apply(
            lambda row: (0.2 <= row['X'] < 0.6) & (0.7 <= row['Y'] < 0.9), axis=1)]


class TestHypercube(AbstractTestHypercube):

    def test_get_dimension(self):
        self.assertEqual(self.dimensions, self.cube.dimensions)

    def test_get_limit_count(self):
        self.assertEqual(0, self.cube.limit_count)
        self.cube.add_limit('X', '+')
        self.assertEqual(1, self.cube.limit_count)
        self.cube.add_limit('Y', '-')
        self.assertEqual(2, self.cube.limit_count)
        self.cube.add_limit('X', '+')
        self.assertEqual(2, self.cube.limit_count)

    def test_get_mean(self):
        self.assertEqual(self.mean, self.cube.output)

    def test_get(self):
        self.assertEqual((0.2, 0.6), self.cube['X'])
        self.assertEqual((0.7, 0.9), self.cube['Y'])
        with self.assertRaises(FeatureNotFoundException):
            dummy = self.cube['Z']

    def test_get_first(self):
        self.assertEqual(0.2, self.cube.get_first('X'))
        self.assertEqual(0.7, self.cube.get_first('Y'))
        with self.assertRaises(FeatureNotFoundException):
            self.cube.get_first('Z')

    def test_get_second(self):
        self.assertEqual(0.6, self.cube.get_second('X'))
        self.assertEqual(0.9, self.cube.get_second('Y'))
        with self.assertRaises(FeatureNotFoundException):
            self.cube.get_second('Z')

    def test_copy(self):
        copy = self.cube.copy()
        self.assertEqual(self.cube.dimensions, copy.dimensions)
        self.assertEqual(self.cube.output, copy.output)

    def test_expand(self):
        arguments = TestHypercube.expansion_provider()
        for arg in arguments:
            arg[0].expand(arg[1], self.hypercubes)
            self.assertEqual(arg[2], arg[0][arg[1].feature])

    def test_expand_all(self):
        updates = [MinUpdate('X', 0.2), MinUpdate('Y', 0.15)]
        surrounding = HyperCube({'X': (0.0, 0.8), 'Y': (0.1, 0.6)}, output=0.0)
        cube = HyperCube({'X': (0.1, 0.2), 'Y': (0.4, 0.4)}, output=0.4)
        cube.expand_all(updates, surrounding)
        self.assertEqual((0.0, 0.4), cube.dimensions['X'])
        self.assertEqual((0.25, 0.55), cube.dimensions['Y'])

    def test_overlap(self):
        self.assertIsNone(self.cube.overlap(self.hypercubes))
        self.assertFalse(self.cube.overlap(self.hypercubes[0]))
        self.assertFalse(self.cube.overlap(self.hypercubes[1]))
        self.cube.update_dimension('X', 0.6, 1.0)
        self.assertIsNotNone(self.cube.overlap(self.hypercubes))
        self.assertEqual(self.hypercubes[1], self.cube.overlap(self.hypercubes))
        self.assertFalse(self.cube.overlap(self.hypercubes[0]))
        self.assertTrue(self.cube.overlap(self.hypercubes[1]))
        self.assertTrue(self.hypercubes[1].overlap(self.cube))

    def test_has_volume(self):
        self.assertTrue(self.cube.has_volume())
        no_volume = self.cube.copy()
        no_volume.update_dimension('X', 1.0, 1.0)
        self.assertFalse(no_volume.has_volume())

    def test_equal(self):
        self.assertTrue(self.cube == self.cube)
        self.assertFalse(self.cube == self.hypercubes[1])

    def test_contains(self):
        arguments = TestHypercube.tuple_provider()
        for arg in arguments:
            self.assertEqual(arg[1], arg[0] in self.cube)

    def test_count(self):
        self.assertEqual(self.dataset.shape[0], HyperCube.create_surrounding_cube(self.dataset).count(self.dataset))
        self.assertEqual(self.filtered_dataset.shape[0], self.cube.count(self.dataset))

    def test_create_samples(self):
        points = self.cube.create_samples(25)
        for k, v in self.cube.dimensions.items():
            self.assertTrue(all((points.loc[:, k] >= v[0]).values))
            self.assertTrue(all((points.loc[:, k] < v[1]).values))

    def test_add_limit(self):
        self.assertEqual(0, self.cube.limit_count)
        self.cube.add_limit('X', '-')
        self.assertEqual(1, self.cube.limit_count)
        self.cube.add_limit('X', '-')
        self.assertEqual(1, self.cube.limit_count)
        self.cube.add_limit('X', '+')
        self.assertEqual(2, self.cube.limit_count)
        self.cube.add_limit('X', '+')
        self.assertEqual(2, self.cube.limit_count)

    def test_check_limits(self):
        self.assertIsNone(self.cube.check_limits('X'))
        self.cube.add_limit('X', '-')
        self.assertEqual('-', self.cube.check_limits('X'))
        self.cube.add_limit('X', '+')
        self.assertEqual('*', self.cube.check_limits('X'))
        self.assertIsNone(self.cube.check_limits('Y'))
        self.cube.add_limit('Y', '+')
        self.assertEqual('+', self.cube.check_limits('Y'))
        self.cube.add_limit('Y', '-')
        self.assertEqual('*', self.cube.check_limits('Y'))

    def test_update_mean(self):
        model = KNeighborsRegressor()
        model.fit(self.dataset.iloc[:, :-1], self.dataset.iloc[:, -1])
        predictor = Predictor(model)
        self.cube.update(self.dataset, predictor)

    def test_update_dimension(self):
        new_lower, new_upper = 0.6, 1.4
        updated = {'X': (new_lower, new_upper),
                   'Y': (0.7, 0.9)}
        new_cube1 = self.cube.copy()
        new_cube1.update_dimension('X', new_lower, new_upper)
        self.assertEqual(updated, new_cube1.dimensions)
        new_cube2 = self.cube.copy()
        new_cube2.update_dimension('X', (new_lower, new_upper))
        self.assertEqual(updated, new_cube2.dimensions)

    def test_create_surrounding_cube(self):
        surrounding = HyperCube.create_surrounding_cube(self.dataset)
        for feature in self.dataset.columns[:-1]:
            self.assertEqual((round(min(self.dataset[feature]) - HyperCube.EPSILON * 2, get_int_precision()),
                              round(max(self.dataset[feature]) + HyperCube.EPSILON * 2, get_int_precision())),
                             surrounding.dimensions[feature])

    def test_cube_from_point(self):
        lower, upper, mean = 0.5, 0.8, 0.6
        cube = HyperCube.cube_from_point({'X': lower, 'Y': upper, 'z': mean})
        self.assertEqual({'X': (lower, lower), 'Y': (upper, upper)}, cube.dimensions)
        self.assertEqual(mean, cube.output)

    def test_check_overlap(self):
        self.assertTrue(HyperCube.check_overlap((self.hypercubes[0], ), (self.hypercubes[0].copy(), )))
        self.assertTrue(HyperCube.check_overlap(self.hypercubes, self.hypercubes + [self.hypercubes[0].copy()]))
        self.assertFalse(HyperCube.check_overlap(self.hypercubes, self.hypercubes))
        self.assertFalse(HyperCube.check_overlap(self.hypercubes[0:1], self.hypercubes[1:]))

    @staticmethod
    def expansion_provider():
        cube1 = HyperCube({'X': (2.3, 6.4), 'Y': (8.9, 12.3)}, output=2.3)
        fake1 = cube1.copy()
        fake1.update_dimension('X', 0.5, 2.3)
        fake2 = cube1.copy()
        fake2.update_dimension('X', 6.4, 12.9)
        cube2 = cube1.copy()
        cube2.update_dimension('X', 9.5, 12.3)
        fake3 = cube2.copy()
        fake3.update_dimension('X', 5.0, 9.5)
        fake4 = cube2.copy()
        fake4.update_dimension('X', 12.3, 15.2)

        return [(cube1.copy(), Expansion(fake1, 'X', '-', 0.0), (0.5, 6.4)),
                (cube1.copy(), Expansion(fake2, 'X', '+', 0.0), (2.3, 6.6)),
                (cube2.copy(), Expansion(fake3, 'X', '-', 0.0), (7.0, 12.3)),
                (cube2.copy(), Expansion(fake4, 'X', '+', 0.0), (9.5, 15.2))]

    @staticmethod
    def tuple_provider():
        return (({'X': 0.5, 'Y': 0.8}, True),
                ({'X': 0.1, 'Y': 0.8}, False),
                ({'X': 0.5, 'Y': 0.95}, False),
                ({'X': 0.1, 'Y': 0.95}, False))


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from statistics import mode
from functools import reduce
from typing import Iterable, Union
import pandas as pd
from numpy import ndarray

from psyke.extraction.hypercubic.utils import Dimension, Dimensions, MinUpdate, ZippedDimension, Limit, Expansion
from psyke.schema import Between
from psyke.utils import get_default_precision, get_int_precision, Target, get_default_random_seed
from psyke.utils.logic import create_term, to_rounded_real, linear_function_creator
from sklearn.linear_model import LinearRegression
from tuprolog.core import Var, Struct
from random import Random
import numpy as np


class FeatureNotFoundException(Exception):

    def __init__(self, feature: str):
        super().__init__('Feature "' + feature + '" not found.')


class HyperCube:
    """
    A N-dimensional cube holding a numeric value.
    """

    EPSILON = get_default_precision()  # Precision used when comparing two hypercubes
    INT_PRECISION = get_int_precision()

    def __init__(self, dimension: dict[str, tuple[float, float]] = None, limits: set[Limit] = None,
                 output: float | LinearRegression = 0.0):
        self._dimensions = self._fit_dimension(dimension) if dimension is not None else {}
        self._limits = limits if limits is not None else set()
        self._output = output
        self._diversity = 0.0

    def __contains__(self, point: dict[str, float]) -> bool:
        """
        Note that a point (dict[str, float]) is inside a hypercube if ALL its dimensions' values satisfy:
            min_dim <= value < max_dim
        :param point: an N-dimensional point
        :return: true if the point is inside the hypercube, false otherwise
        """
        return all([(self.get_first(k) <= v < self.get_second(k)) for k, v in point.items()])

    def __eq__(self, other: HyperCube) -> bool:
        return all([(abs(dimension.this_dimension[0] - dimension.other_dimension[0]) < HyperCube.EPSILON)
                    & (abs(dimension.this_dimension[1] - dimension.other_dimension[1]) < HyperCube.EPSILON)
                    for dimension in self._zip_dimensions(other)])

    def __getitem__(self, feature: str) -> Dimension:
        if feature in self._dimensions.keys():
            return self._dimensions[feature]
        else:
            raise FeatureNotFoundException(feature)

    def __setitem__(self, key, value) -> None:
        self._dimensions[key] = value

    def __hash__(self) -> int:
        result = [hash(name + str(dimension[0]) + str(dimension[1])) for name, dimension in self.dimensions.items()]
        return sum(result)

    @property
    def dimensions(self) -> Dimensions:
        return self._dimensions

    @property
    def limit_count(self) -> int:
        return len(self._limits)

    @property
    def output(self) -> float | LinearRegression:
        return self._output

    @property
    def diversity(self) -> float:
        return self._diversity

    def _fit_dimension(self, dimension: dict[str, tuple[float, float]]) -> dict[str, tuple[float, float]]:
        new_dimension: dict[str, tuple[float, float]] = {}
        for key, value in dimension.items():
            new_dimension[key] = (round(value[0], self.INT_PRECISION), round(value[1], self.INT_PRECISION))
        return new_dimension

    def _expand_one(self, update: MinUpdate, surrounding: HyperCube, ratio: float = 1.0) -> None:
        self.update_dimension(update.name, (
            max(self.get_first(update.name) - update.value / ratio, surrounding.get_first(update.name)),
            min(self.get_second(update.name) + update.value / ratio, surrounding.get_second(update.name))
        ))

    def filter_indices(self, dataset: pd.DataFrame) -> ndarray:
        v = np.array([v for _, v in self._dimensions.items()])
        ds = dataset.to_numpy(copy=True)
        return np.all((v[:, 0] <= ds) & (ds < v[:, 1]), axis=1)

    def _filter_dataframe(self, dataset: pd.DataFrame) -> pd.DataFrame:
        return dataset[self.filter_indices(dataset)]

    def _zip_dimensions(self, hypercube: HyperCube) -> list[ZippedDimension]:
        return [ZippedDimension(dimension, self[dimension], hypercube[dimension])
                for dimension in self._dimensions.keys()]

    def add_limit(self, limit_or_feature: Limit | str, direction: str = None) -> None:
        if isinstance(limit_or_feature, Limit):
            self._limits.add(limit_or_feature)
        else:
            self.add_limit(Limit(limit_or_feature, direction))

    def check_limits(self, feature: str) -> str | None:
        filtered = [limit for limit in self._limits if limit.feature == feature]
        if len(filtered) == 0:
            return None
        if len(filtered) == 1:
            return filtered[0].direction
        if len(filtered) == 2:
            return '*'
        raise Exception('Too many limits for this feature')

    def create_samples(self, n: int = 1, generator: Random = Random(get_default_random_seed())) -> pd.DataFrame:
        return pd.DataFrame([self._create_tuple(generator) for _ in range(n)])

    @staticmethod
    def check_overlap(to_check: Iterable[HyperCube], hypercubes: Iterable[HyperCube]) -> bool:
        checked = []
        to_check_copy = list(to_check).copy()
        while len(to_check_copy) > 0:
            cube = to_check_copy.pop()
            for hypercube in hypercubes:
                if hypercube not in checked and cube.overlap(hypercube):
                    return True
            checked += [cube]
        return False

    def copy(self) -> HyperCube:
        return HyperCube(self.dimensions.copy(), self._limits.copy(), self.output)

    def count(self, dataset: pd.DataFrame) -> int:
        return self._filter_dataframe(dataset.iloc[:, :-1]).shape[0]

    def body(self, variables: dict[str, Var], ignore: list[str], unscale=None, normalization=None) -> Iterable[Struct]:
        dimensions = dict(self.dimensions)
        for dimension in ignore:
            del dimensions[dimension]
        return [create_term(variables[name], Between(unscale(values[0], name), unscale(values[1], name)))
                for name, values in dimensions.items()]

    @staticmethod
    def create_surrounding_cube(dataset: pd.DataFrame, closed: bool = False,
                                output=None) -> GenericCube:
        output = Target.CONSTANT if output is None else output
        dimensions = {
            column: (min(dataset[column]) - HyperCube.EPSILON * 2, max(dataset[column]) + HyperCube.EPSILON * 2)
            for column in dataset.columns[:-1]
        }
        if closed:
            if output == Target.CONSTANT:
                return ClosedCube(dimensions)
            if output == Target.REGRESSION:
                return ClosedRegressionCube(dimensions)
            return ClosedClassificationCube(dimensions)
        if output == Target.CLASSIFICATION:
            return ClassificationCube(dimensions)
        if output == Target.REGRESSION:
            return RegressionCube(dimensions)
        return HyperCube(dimensions)

    def _create_tuple(self, generator: Random) -> dict:
        return {k: generator.uniform(self.get_first(k), self.get_second(k)) for k in self._dimensions.keys()}

    @staticmethod
    def cube_from_point(point: dict, output=None) -> GenericCube:
        if output is Target.CLASSIFICATION:
            return ClassificationCube({k: (v, v) for k, v in list(point.items())[:-1]})
        if output is Target.REGRESSION:
            return RegressionCube({k: (v, v) for k, v in list(point.items())[:-1]})
        return HyperCube({k: (v, v) for k, v in list(point.items())[:-1]}, output=list(point.values())[-1])

    def equal(self, hypercubes: Iterable[HyperCube] | HyperCube) -> bool:
        if isinstance(hypercubes, Iterable):
            return any([self.equal(cube) for cube in hypercubes])
        else:
            return all([(abs(dimension.this_dimension[0] - dimension.other_dimension[0]) < HyperCube.EPSILON)
                        & (abs(dimension.this_dimension[1] - dimension.other_dimension[1]) < HyperCube.EPSILON)
                        for dimension in self._zip_dimensions(hypercubes)])

    def expand(self, expansion: Expansion, hypercubes: Iterable[HyperCube]) -> None:
        feature = expansion.feature
        a, b = self[feature]
        self.update_dimension(feature, expansion.boundaries(a, b))
        other_cube = self.overlap(hypercubes)
        if isinstance(other_cube, HyperCube):
            self.update_dimension(feature, (other_cube.get_second(feature), b)
            if expansion.direction == '-' else (a, other_cube.get_first(feature)))
        if isinstance(self.overlap(hypercubes), HyperCube):
            raise Exception('Overlapping not handled')

    def expand_all(self, updates: Iterable[MinUpdate], surrounding: HyperCube, ratio: float = 1.0) -> None:
        for update in updates:
            self._expand_one(update, surrounding, ratio)

    def get_first(self, feature: str) -> float:
        return self[feature][0]

    def get_second(self, feature: str) -> float:
        return self[feature][1]

    def has_volume(self) -> bool:
        return all([dimension[1] - dimension[0] > HyperCube.EPSILON for dimension in self._dimensions.values()])

    def volume(self) -> float:
        return reduce(lambda a, b: a * b, [dimension[1] - dimension[0] for dimension in self._dimensions.values()], 1)

    def diagonal(self) -> float:
        return reduce(
            lambda a, b: a + b, [(dimension[1] - dimension[0]) ** 2 for dimension in self._dimensions.values()], 0
        ) ** 0.5

    def is_adjacent(self, cube: HyperCube) -> str | None:
        adjacent = None
        for (feature, [a1, b1]) in self._dimensions.items():
            if self[feature] == cube[feature]:
                continue
            [a2, b2] = cube[feature]
            if (adjacent is not None) or ((b1 != a2) and (b2 != a1)):
                return None
            adjacent = feature
        return adjacent

    def merge_along_dimension(self, cube: HyperCube, feature: str) -> HyperCube:
        new_cube = self.copy()
        (a1, b1) = self[feature]
        (a2, b2) = cube[feature]
        new_cube.update_dimension(feature, (min(a1, a2), max(b1, b2)))
        return new_cube

    # TODO: maybe two different methods are more readable and easier to debug
    def overlap(self, hypercubes: Iterable[HyperCube] | HyperCube) -> HyperCube | bool | None:
        if isinstance(hypercubes, Iterable):
            for hypercube in hypercubes:
                if (self != hypercube) & self.overlap(hypercube):
                    return hypercube
            return None
        elif self is hypercubes:
            return False
        else:
            return all([not ((dimension.other_dimension[0] >= dimension.this_dimension[1]) |
                             (dimension.this_dimension[0] >= dimension.other_dimension[1]))
                        for dimension in self._zip_dimensions(hypercubes)])

    # TODO: maybe two different methods are more readable and easier to debug
    def update_dimension(self, feature: str, lower: float | tuple[float, float], upper: float | None = None) -> None:
        if upper is None:
            self[feature] = lower
        else:
            self.update_dimension(feature, (lower, upper))

    def update(self, dataset: pd.DataFrame, predictor) -> None:
        filtered = self._filter_dataframe(dataset.iloc[:, :-1])
        predictions = predictor.predict(filtered)
        self._output = np.mean(predictions)
        self._diversity = np.std(predictions)

    # TODO: why this is not a property?
    def init_std(self, std: float) -> None:
        self._diversity = std


class RegressionCube(HyperCube):
    def __init__(self, dimension: dict[str, tuple] = None):
        super().__init__(dimension=dimension, output=LinearRegression())

    def update(self, dataset: pd.DataFrame, predictor) -> None:
        filtered = self._filter_dataframe(dataset.iloc[:, :-1])
        if len(filtered > 0):
            predictions = predictor.predict(filtered)
            self._output.fit(filtered, predictions)
            self._diversity = (abs(self._output.predict(filtered) - predictions)).mean()

    def copy(self) -> RegressionCube:
        return RegressionCube(self.dimensions.copy())

    def body(self, variables: dict[str, Var], ignore: list[str], unscale=None, normalization=None) -> Iterable[Struct]:
        intercept = self.output.intercept_ if normalization is None else \
            unscale(sum([-self.output.coef_[i] * normalization[name][0] / normalization[name][1] for i, name in
                         enumerate(self.dimensions.keys())], self.output.intercept_), list(normalization.keys())[-1])
        coefs = self.output.coef_ if normalization is None else \
            [self.output.coef_[i] / normalization[name][1] for i, name in enumerate(self.dimensions.keys())]
        return list(super().body(variables, ignore, unscale, normalization)) + [linear_function_creator(
            list(variables.values()), [to_rounded_real(v) for v in coefs],
            to_rounded_real(intercept)
        )]


class ClassificationCube(HyperCube):
    def __init__(self, dimension: dict[str, tuple] = None):
        super().__init__(dimension=dimension)

    def update(self, dataset: pd.DataFrame, predictor) -> None:
        filtered = self._filter_dataframe(dataset.iloc[:, :-1])
        if len(filtered > 0):
            predictions = predictor.predict(filtered)
            self._output = mode(predictions)
            self._diversity = 1 - sum(prediction == self.output for prediction in predictions) / len(filtered)

    def copy(self) -> ClassificationCube:
        return ClassificationCube(self.dimensions.copy())


class ClosedCube(HyperCube):
    def __init__(self, dimension: dict[str, tuple] = None):
        super().__init__(dimension=dimension)

    def __contains__(self, point: dict[str, float]) -> bool:
        return all([(self.get_first(k) <= v <= self.get_second(k)) for k, v in point.items()])

    def filter_indices(self, dataset: pd.DataFrame) -> ndarray:
        v = np.array([v for _, v in self._dimensions.items()])
        ds = dataset.to_numpy(copy=True)
        return np.all((v[:, 0] <= ds) & (ds <= v[:, 1]), axis=1)

    def copy(self) -> ClosedCube:
        return ClosedCube(self.dimensions.copy())


class ClosedRegressionCube(ClosedCube, RegressionCube):
    def __init__(self, dimension: dict[str, tuple] = None):
        super().__init__(dimension=dimension)

    def copy(self) -> ClosedRegressionCube:
        return ClosedRegressionCube(self.dimensions.copy())


class ClosedClassificationCube(ClosedCube, ClassificationCube):
    def __init__(self, dimension: dict[str, tuple] = None):
        super().__init__(dimension=dimension)

    def copy(self) -> ClosedClassificationCube:
        return ClosedClassificationCube(self.dimensions.copy())


GenericCube = Union[HyperCube, ClassificationCube, RegressionCube,
                    ClosedCube, ClosedRegressionCube, ClosedClassificationCube]



import unittest
from psyke.extraction.real.utils import Rule
from psyke.utils.dataframe import split_features
from test import get_dataset


class TestRule(unittest.TestCase):

    def test_subrule(self):
        pred_1, pred_2 = ['V1', 'V2'], ['V3', 'V4']
        rule_1 = Rule(pred_1, pred_2)
        self.assertTrue(rule_1 in rule_1)
        rule_2 = Rule(pred_2, pred_1)
        self.assertFalse(rule_1 in rule_2)
        self.assertFalse(rule_2 in rule_1)
        rule_3 = Rule(['V1'], ['V3'])
        self.assertTrue(rule_1 in rule_3)
        self.assertFalse(rule_3 in rule_1)
        self.assertFalse(rule_2 in rule_3)
        self.assertFalse(rule_3 in rule_2)
        rule_4 = Rule(["V1"], ["V5"])
        self.assertFalse(rule_1 in rule_4)
        self.assertFalse(rule_4 in rule_1)
        rule_5 = Rule(["V1", "V6"], ["V3", "V4"])
        self.assertFalse(rule_1 in rule_5)
        self.assertFalse(rule_5 in rule_1)
        self.assertTrue(rule_1 in Rule([], []))

    def test_reduce(self):
        dataset = get_dataset('iris')
        features = split_features(dataset)
        rule = Rule(["V1_1", "V2_2", "V3_0"],
                    ["V1_0", "V2_1", "V2_0", "V4_1", "V4_2"])
        reduced_rule = Rule(["V1_1", "V2_2", "V3_0"],
                            ["V4_1", "V4_2"])
        self.assertEqual(reduced_rule.true_predicates, rule.reduce(features).true_predicates)
        self.assertEqual(reduced_rule.false_predicates, rule.reduce(features).false_predicates)
        self.assertEqual(reduced_rule.true_predicates, reduced_rule.reduce(features).true_predicates)
        self.assertEqual(reduced_rule.false_predicates, reduced_rule.reduce(features).false_predicates)


if __name__ == '__main__':
    unittest.main()


class DatasetError(Exception):
    def __init__(self, message: str = 'Dataset not found'):
        self.message = message
        super().__init__(self.message)


class PredictorError(Exception):
    def __init__(self, message: str = 'Predictor not found'):
        self.message = message
        super().__init__(self.message)


class ExtractorError(Exception):
    def __init__(self, message: str = 'Extractor not found'):
        self.message = message
        super().__init__(self.message)


class SVMError(Exception):
    def __init__(self, message: str = "Kernel must be one of ['linear', 'poly', 'rbf', 'sigmoid', 'precomputed']"):
        self.message = message
        super().__init__(self.message)


TASKS = ['Classification', 'Regression']

DATASETS = [
    ['Iris', [TASKS[0]]],
    ['Wine', [TASKS[0]]],
    ['Arti', [TASKS[1]]],
    ['House', [TASKS[1]]],
    ['Bank marketing', [TASKS[1]]],
    ['Custom', TASKS]
]

FIXED_PREDICTOR_PARAMS = {
    'Test set': (0.5, 'float', None),
    'Split seed': (0, 'int', None)
}

PREDICTORS = {
    'K-NN': [TASKS, {'K': (3, 'int', None)}],
    'DT': [TASKS, {'Max depth': (3, 'int', None), 'Max leaves': (3, 'int', None)}],
    'RF': [TASKS, {'N estimators': (50, 'int', None), 'Max depth': (3, 'int', None), 'Max leaves': (100, 'int', None)}],
    'LR': [[TASKS[1]], {}],
    'SVM': [TASKS, {'Regularization': (1.0, 'float', None),
                    'Kernel': ('RBF', ['Linear', 'Poly', 'RBF', 'Sigmoid'], None),
                    'Epsilon': (0.1, 'float', TASKS[1])}]
}

MAX_DEPTH = (3, 'int', None)
THRESHOLD = (0.1, 'float', None)
MIN_EXAMPLES = (100, 'int', None)

EXTRACTORS = {
    'REAL': [[TASKS[0]], {}],
    'Trepan': [[TASKS[0]], {'Max depth': MAX_DEPTH, 'Min examples': (0, 'int', None)}],
    'CART': [TASKS, {'Max depth': MAX_DEPTH, 'Max leaves': (3, 'int', None), 'Simplify': (True, 'bool', None)}],
    'Iter': [TASKS, {'Min examples': MIN_EXAMPLES, 'Threshold': THRESHOLD, 'Max iterations': (600, 'int', None),
                     'N points': (1, 'int', None), 'Min update': (0.05, 'float', None),
                     'Fill gaps': (True, 'bool', None)}],
    'GridEx': [TASKS, {'Max depth': MAX_DEPTH, 'Splits': (2, 'int', None), 'Adaptive': (True, 'bool', None),
                       'Adaptive threshold': (0.8, 'float', None), 'Min examples': MIN_EXAMPLES,
                       'Threshold': THRESHOLD}],
    'GridREx': [[TASKS[1]], {'Max depth': MAX_DEPTH, 'Splits': (2, 'int', None), 'Adaptive': (True, 'bool', None),
                             'Adaptive threshold': (0.8, 'float', None), 'Min examples': MIN_EXAMPLES,
                             'Threshold': THRESHOLD}],
    'CReEPy': [TASKS, {'Max depth': MAX_DEPTH, 'Threshold': THRESHOLD, 'Max components': (10, 'int', None),
                       'Feat threshold': (0.8, 'float', None), 'Constant output': (True, 'bool', TASKS[1])}],
    'ORCHiD': [TASKS, {'Max depth': MAX_DEPTH, 'Threshold': THRESHOLD, 'Max components': (10, 'int', None),
                       'Feat threshold': (0.8, 'float', None), 'Constant output': (True, 'bool', TASKS[1])}]
}


def cast_param(params, key, value):
    if params[key][1] is None:
        return value
    if params[key][1] == 'int':
        return int(value)
    if params[key][1] == 'float':
        return float(value)
    if params[key][1] == 'bool':
        return bool(value)
    if isinstance(params[key][1], list):
        if value not in params[key][1]:
            raise ValueError
        return value
    raise NotImplementedError



from functools import lru_cache
from psyke.extraction.real.utils import Rule, IndexedRuleSet
from psyke import PedagogicalExtractor
from psyke.schema import DiscreteFeature
from psyke.utils.dataframe import HashableDataFrame
from psyke.utils.logic import create_term, create_head, create_variable_list
from tuprolog.core import Var, Struct, Clause, clause
from tuprolog.theory import MutableTheory, mutable_theory, Theory
from typing import Iterable
import pandas as pd
import numpy as np


class REAL(PedagogicalExtractor):
    """
    Explanator implementing Rule Extraction As Learning (REAL) algorithm, doi:10.1016/B978-1-55860-335-6.50013-1.
    The algorithm is sensible to features' order in the provided dataset during extraction.
    To make it reproducible the features are internally sorted (alphabetically).
    """

    def __init__(self, predictor, discretization: Iterable[DiscreteFeature]):
        super().__init__(predictor, discretization)
        self._ruleset: IndexedRuleSet = IndexedRuleSet()
        self._output_mapping = {}

    @property
    def n_rules(self):
        return len(self._ruleset.flatten())

    def _covers(self, sample: pd.Series, rules: list[Rule]) -> bool:
        new_rule = self._rule_from_example(sample)
        return any([new_rule in rule for rule in rules])

    def _create_body(self, variables: dict[str, Var], rule: Rule) -> list[Struct]:
        result = []
        for predicates, truth_value in zip(rule.to_lists(), [True, False]):
            for predicate in predicates:
                feature = [feature for feature in self.discretization if predicate in feature.admissible_values][0]
                result.append(create_term(variables[feature.name], feature.admissible_values[predicate], truth_value))
        return result

    def _create_clause(self, dataset: pd.DataFrame, variables: dict[str, Var], key: int, rule: Rule) -> Clause:
        head = create_head(dataset.columns[-1],
                           sorted(list(variables.values())),
                           str(sorted(list(set(dataset.iloc[:, -1])))[key]))
        return clause(head, self._create_body(variables, rule))

    def _create_new_rule(self, sample: pd.Series) -> Rule:
        rule = self._rule_from_example(sample)
        return self._generalise(rule, sample)

    def _create_ruleset(self, dataset: pd.DataFrame) -> IndexedRuleSet:
        ruleset = IndexedRuleSet.create_indexed_ruleset(dataset)
        for index, sample in dataset.iloc[:, :-1].iterrows():
            prediction = list(self.predictor.predict(sample.to_frame().transpose()))[0]
            rules = ruleset.get(self._output_mapping[prediction])
            if not self._covers(sample, rules):
                rules.append(self._create_new_rule(sample))
        return ruleset.optimize()

    def _create_theory(self, dataset: pd.DataFrame, ruleset: IndexedRuleSet, sort: bool = True) -> MutableTheory:
        theory = mutable_theory()
        for key, rule in ruleset.flatten():
            variables = create_variable_list(self.discretization, sort=sort)
            theory.assertZ(self._create_clause(dataset, variables, key, rule))
        return theory

    def _generalise(self, rule: Rule, sample: pd.Series) -> Rule:
        mutable_rule = rule.to_lists()
        samples = sample.to_frame().transpose()
        for predicate in rule.true_predicates:
            samples = self._remove_antecedent(samples.copy(), predicate, mutable_rule)
        return Rule(mutable_rule[0], mutable_rule[1]).reduce(self.discretization)

    def _remove_antecedent(self, samples: pd.DataFrame, predicate: str, rule: list[list[str]]) -> (pd.DataFrame, bool):
        feature = [feature for feature in self.discretization if predicate in feature.admissible_values][0]
        output = np.array(self.predictor.predict(samples))
        copies = [samples.copy()]
        samples[predicate] = 0
        for f in [f for f in feature.admissible_values if f != predicate]:
            copy = samples.copy()
            copy[f] = 1
            if all(output == np.array(self.predictor.predict(copy))):
                copies.append(copy)
                rule[1].remove(f)
        if len(copies) > 1:
            rule[0].remove(predicate)
        return pd.concat([df for df in copies], ignore_index=True)

    @lru_cache(maxsize=512)
    def _get_or_set(self, dataset: HashableDataFrame) -> IndexedRuleSet:
        return self._create_ruleset(dataset)

    def _internal_predict(self, sample: pd.Series):
        x = [index for index, rule in self._ruleset.flatten() if REAL._rule_from_example(sample) in rule]
        reverse_mapping = dict((v, k) for k, v in self._output_mapping.items())
        return reverse_mapping[x[0]] if len(x) > 0 else None

    @staticmethod
    def _rule_from_example(sample: pd.Series) -> Rule:
        true_predicates, false_predicates = [], []
        for feature, value in sample.items():
            true_predicates.append(str(feature)) if value == 1 else false_predicates.append(str(feature))
        return Rule(sorted(true_predicates), sorted(false_predicates))

    def _subset(self, samples: pd.DataFrame, predicate: str) -> (pd.DataFrame, bool):
        samples_0 = samples.copy()
        samples_0[predicate].values[:] = 0
        samples_1 = samples.copy()
        samples_1[predicate].values[:] = 1
        samples_all = samples_0.append(samples_1)
        return samples_all, len(set(self.predictor.predict(samples_all))) == 1

    def _extract(self, dataframe: pd.DataFrame, mapping: dict[str: int] = None, sort: bool = True) -> Theory:
        # Order the dataset by column to preserve reproducibility.
        dataframe = dataframe.sort_values(by=list(dataframe.columns.values), ascending=False)
        # Always perform output mapping in the same (sorted) way to preserve reproducibility.
        if mapping is None:
            self._output_mapping = {value: index for index, value in enumerate(sorted(set(dataframe.iloc[:, -1])))}
        else:
            self._output_mapping = {value: index for index, value in enumerate(sorted(set(mapping[dataframe.iloc[:, -1]])))}
        self._ruleset = self._get_or_set(HashableDataFrame(dataframe))
        return self._create_theory(dataframe, self._ruleset, sort)

    def _predict(self, dataframe) -> Iterable:
        return np.array([self._internal_predict(data.transpose()) for _, data in dataframe.iterrows()])



from collections import Iterable
from typing import Union, Any
import numpy as np
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from psyke.schema import Value, LessThan, GreaterThan

LeafConstraints = list[tuple[str, Value, bool]]
LeafSequence = Iterable[tuple[LeafConstraints, Any]]


class CartPredictor:
    """
    A wrapper for decision and regression trees of sklearn.
    """

    def __init__(self, predictor: Union[DecisionTreeClassifier, DecisionTreeRegressor] = DecisionTreeClassifier(),
                 normalization=None):
        self._predictor = predictor
        self.normalization = normalization

    def __get_constraints(self, nodes: Iterable[(int, bool)]) -> LeafConstraints:
        thresholds = [self._predictor.tree_.threshold[i[0]] for i in nodes]
        features = [self._predictor.feature_names_in_[self._predictor.tree_.feature[node[0]]] for node in nodes]
        conditions = [node[1] for node in nodes]
        if self.normalization is not None:
            thresholds = [threshold * self.normalization[feature][1] + self.normalization[feature][0]
                          for feature, threshold in zip(features, thresholds)]
        return [(feature, LessThan(threshold) if condition else GreaterThan(threshold), condition)
                for feature, condition, threshold in zip(features, conditions, thresholds)]

    def __get_leaves(self) -> Iterable[int]:
        return [i for i, (left_child, right_child) in enumerate(zip(
            self._left_children, self._right_children
        )) if left_child == -1 and right_child == -1]

    def __get_prediction(self, node: int) -> Any:
        if hasattr(self._predictor, 'classes_'):
            return self._predictor.classes_[np.argmax(self._predictor.tree_.value[node])]
        else:
            return self._predictor.tree_.value[node]

    def __path(self, node: int, path=[]) -> Iterable[(int, bool)]:
        if node == 0:
            return path
        father = list(self._left_children if node in self._left_children else self._right_children).index(node)
        return self.__path(father, [(father, node in self._left_children)] + path)

    def __iter__(self) -> LeafSequence:
        leaves = self.__get_leaves()
        return ((self.__get_constraints(self.__path(i)), self.__get_prediction(i)) for i in leaves)

    def predict(self, data) -> Iterable:
        return self._predictor.predict(data)

    @property
    def predictor(self) -> Union[DecisionTreeClassifier, DecisionTreeRegressor]:
        return self._predictor

    @property
    def n_leaves(self) -> int:
        return len(list(self.__get_leaves()))

    @property
    def _left_children(self) -> list[int]:
        return self._predictor.tree_.children_left

    @property
    def _right_children(self) -> list[int]:
        return self._predictor.tree_.children_right

    @predictor.setter
    def predictor(self, predictor: Union[DecisionTreeClassifier, DecisionTreeRegressor]):
        self._predictor = predictor



from parameterized import parameterized_class
from psyke.utils import get_default_precision
from psyke import logger
from test.psyke import initialize
import unittest

""" 
    TODO (?): right now there is a small chance that corner data are wrongly predicted (that is fine for now).
    In other words, if we use the extracted rules (with a specific default accuracy fo float)
    and compare their result with the one obtained by the actual decision tree (thresholds do not have truncated float)
    they may be different. To avoid this, when we will refactor all extractor we will also address this issue.
"""


@parameterized_class(initialize('cart'))
class TestCart(unittest.TestCase):

    def test_extract(self):
        logger.info(self.expected_theory)
        logger.info(self.extracted_theory)
        self.assertTrue(self.expected_theory.equals(self.extracted_theory, False))

    def test_predict(self):
        # self.assertEqual(self.extracted_test_y_from_theory, self.extracted_test_y_from_pruned_theory)
        if isinstance(self.extracted_test_y_from_theory[0], str):
            self.assertTrue(all(self.extracted_test_y_from_theory == self.extracted_test_y_from_extractor))
        else:
            self.assertTrue(max(abs(self.extracted_test_y_from_theory - self.extracted_test_y_from_extractor)) <
                            get_default_precision())


if __name__ == '__main__':
    unittest.main()



from pathlib import Path
from typing import TextIO

PATH = Path(__file__).parents[0]


def get_predictor_path(filename: str) -> Path:
    return PATH / f"{filename}.onnx"


def open_predictor(filename: str) -> TextIO:
    return open(get_predictor_path(filename))


def create_predictor_name(dataset: str, model_type: str, options: dict) -> str:
    options = ''.join('_' + k + '-' + str(v) for k, v in options.items())
    return dataset + model_type + options



from __future__ import annotations

from typing import Iterable

import numpy as np
import pandas as pd

import psyke.utils
from psyke.clustering.exact import ExACT
from psyke.extraction.hypercubic import Node, HyperCube, ClosedCube
from psyke.clustering.utils import select_gaussian_mixture


class CREAM(ExACT):
    """
    Explanator implementing CREAM algorithm.
    """

    def __init__(self, depth: int, error_threshold: float,
                 output: psyke.utils.Target = psyke.utils.Target.CONSTANT, gauss_components: int = 5):
        super().__init__(depth, error_threshold, output, gauss_components)

    def __eligible_cubes(self, gauss_pred: np.ndarray, node: Node, clusters: int):
        cubes = []
        for i in range(len(np.unique(gauss_pred))):
            df = node.dataframe.iloc[np.where(gauss_pred == i)]
            if len(df) == 0:
                continue
            inner_cube = self._create_cube(df, clusters)
            indices = self._indices(inner_cube, node.dataframe)
            if indices is None:
                continue
            right, left = self._split(inner_cube, node.cube, node.dataframe, indices)
            cubes.append((
                ((right.diversity + left.diversity) / 2, right.volume(), left.volume(), i),
                (right, indices), (left, ~indices)
            ))
        return cubes

    def _split(self, right: ClosedCube, outer_cube: ClosedCube, data: pd.DataFrame, indices: np.ndarray):
        right.update(data.iloc[indices], self._predictor)
        left = outer_cube.copy()
        left.update(data.iloc[~indices], self._predictor)
        return right, left

    def _iterate(self, surrounding: Node) -> Iterable[HyperCube]:
        to_split = [(self.error_threshold * 10, 1, 1, surrounding)]
        while len(to_split) > 0:
            to_split.sort(reverse=True)
            (_, depth, _, node) = to_split.pop()
            data = ExACT._remove_string_label(node.dataframe)
            gauss_params = select_gaussian_mixture(data, self.gauss_components)
            gauss_pred = gauss_params[2].predict(data)
            cubes = self.__eligible_cubes(gauss_pred, node, gauss_params[1])
            if len(cubes) < 1:
                continue
            _, right, left = min(cubes)
            node.right = Node(node.dataframe[right[1]], right[0])
            node.cube.update(node.dataframe[left[1]], self._predictor)
            node.left = Node(node.dataframe[left[1]], left[0])

            if depth < self.depth:
                to_split += [
                    (error, depth + 1, np.random.uniform(), n) for (n, error) in
                    zip(node.children, [right[0].diversity, left[0].diversity]) if error > self.error_threshold
                ]
        return self._node_to_cubes(surrounding)



from psyke import get_default_random_seed
from psyke.extraction.hypercubic import Grid, RegressionCube
from psyke.extraction.hypercubic.gridex import GridEx


class GridREx(GridEx):
    """
    Explanator implementing GridREx algorithm.
    """

    def __init__(self, predictor, grid: Grid, min_examples: int, threshold: float, normalization,
                 seed=get_default_random_seed()):
        super().__init__(predictor, grid, min_examples, threshold, normalization, seed)

    def _default_cube(self) -> RegressionCube:
        return RegressionCube()



from psyke.extraction.trepan import Node
from test import get_dataset
import pandas as pd
import unittest


class TestNode(unittest.TestCase):

    dataset: pd.DataFrame = get_dataset('iris')
    n_examples = dataset.shape[0]
    all_node = Node(dataset, n_examples)
    setosa_40 = Node(dataset.iloc[10:70, :], n_examples)
    versicolor_10 = Node(dataset.iloc[95:110, :], n_examples)
    virginica_50 = Node(dataset.iloc[20:130, :], n_examples)

    def test_reach(self):
        node = Node(self.dataset, self.n_examples)
        self.assertEqual(node.reach, self.all_node.reach)
        self.assertTrue(self.versicolor_10.reach < self.setosa_40.reach)
        self.assertTrue(self.setosa_40.reach < self.virginica_50.reach)
        self.assertTrue(self.virginica_50.reach < self.all_node.reach)

    def test_dominant(self):
        self.assertEqual('setosa', self.setosa_40.dominant)
        self.assertEqual('virginica', self.virginica_50.dominant)
        self.assertEqual('versicolor', self.versicolor_10.dominant)

    def test_correct(self):
        self.assertEqual(50, self.virginica_50.correct)
        self.assertEqual(40, self.setosa_40.correct)
        self.assertEqual(10, self.versicolor_10.correct)

    def test_fidelity(self):
        self.assertEqual(50 / 150, self.all_node.fidelity)
        self.assertEqual(40 / 60, self.setosa_40.fidelity)
        self.assertEqual(10 / 15, self.versicolor_10.fidelity)
        self.assertEqual(50 / 110, self.virginica_50.fidelity)

    def test_priority(self):
        self.assertTrue(self.all_node.priority < self.virginica_50.priority)
        self.assertTrue(self.virginica_50.priority < self.setosa_40.priority)
        self.assertTrue(self.setosa_40.priority < self.versicolor_10.priority)

    def test_n_classes(self):
        self.assertEqual(3, self.all_node.n_classes)
        self.assertEqual(2, self.versicolor_10.n_classes)
        self.assertEqual(2, self.setosa_40.n_classes)
        self.assertEqual(3, self.virginica_50.n_classes)
        self.assertEqual(1, Node(self.dataset.iloc[15:40, :], self.n_examples).n_classes)

    def test_iterator(self):
        node = Node(self.dataset, self.n_examples)
        child_1 = Node(self.dataset.iloc[:50, :], self.n_examples)
        child_2 = Node(self.dataset.iloc[50:150, :], self.n_examples)
        node.children = [child_1, child_2]
        grandchild_1_1 = Node(self.dataset.iloc[:25, :], self.n_examples)
        grandchild_2_1 = Node(self.dataset.iloc[50:80, :], self.n_examples)
        grandchild_2_2 = Node(self.dataset.iloc[80:120, :], self.n_examples)
        child_1.children = [grandchild_1_1]
        child_2.children = [grandchild_2_1, grandchild_2_2]
        self.assertEqual(list(node), list(child_1) + list(child_2) + [node])
        self.assertEqual([grandchild_1_1, child_1, grandchild_2_1, grandchild_2_2, child_2, node], list(node))

    def test_to_string(self):
        node = Node(self.dataset, self.n_examples, (('V1', 0.0), ('V2', 1.0)))
        self.assertEqual(' = setosa', str(self.all_node))
        self.assertEqual(' = setosa', str(self.setosa_40))
        self.assertEqual(' = virginica', str(self.virginica_50))
        self.assertEqual(' = versicolor', str(self.versicolor_10))
        self.assertEqual('!V1, V2 = setosa', str(node))


if __name__ == '__main__':
    unittest.main()



from functools import partial

from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score, accuracy_score, f1_score


def mae(expected, predicted):
    """
    Calculates the predictions' MAE w.r.t. the instances given as input.

    :param expected: the expected data .
    :param predicted: the predicted data.
    :return: the mean absolute error (MAE) of the predictions.
    """
    return score(expected, predicted, mean_absolute_error)


def mse(expected, predicted):
    """
    Calculates the predictions' MSE w.r.t. the instances given as input.

    :param expected: the expected data .
    :param predicted: the predicted data.
    :return: the mean squared error (MSE) of the predictions.
    """
    return score(expected, predicted, mean_squared_error)


def r2(expected, predicted):
    """
    Calculates the predictions' R2 w.r.t. the instances given as input.

    :param expected: the expected data .
    :param predicted: the predicted data.
    :return: the R2 score of the predictions.
    """
    return score(expected, predicted, r2_score)


def accuracy(expected, predicted):
    """
    Calculates the predictions' classification accuracy w.r.t. the instances given as input.

    :param expected: the expected data .
    :param predicted: the predicted data.
    :return: the classification accuracy of the predictions.
    """
    return score(expected, predicted, accuracy_score)


def f1(expected, predicted):
    """
    Calculates the predictions' F1 score w.r.t. the instances given as input.

    :param expected: the expected data .
    :param predicted: the predicted data.
    :return: the F1 score of the predictions.
    """
    return score(expected, predicted, partial(f1_score, average='weighted'))


def score(expected, predicted, scoring_function):
    """
    Calculates the predictions' score w.r.t. the instances given as input with the provided scoring function.

    :param expected: the expected data .
    :param predicted: the predicted data.
    :param scoring_function: the scoring function to be used.
    :return: the score of the predictions.
    """
    idx = [prediction is not None for prediction in predicted]
    return scoring_function(expected[idx], predicted[idx])



from kivy.graphics import Color, Rectangle, Line, RoundedRectangle
from kivy.uix.relativelayout import RelativeLayout
from kivy.uix.screenmanager import Screen

from psyke.gui.view.DataPanel import DataPanel
from psyke.gui.view.ExtractorPanel import ExtractorPanel
from psyke.gui.view.FeaturePanel import FeaturePanel
from psyke.gui.view.PlotPanel import PlotPanel
from psyke.gui.view.PredictorPanel import PredictorPanel
from psyke.gui.view.TheoryPanel import TheoryPanel


class View(Screen):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class MainScreen(View):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.data_panel = None
        self.feature_panel = None
        self.predictor_panel = None
        self.plot_panel = None
        self.extractor_panel = None
        self.theory_panel = None

        self.l = .33
        self.r = 1 - self.l
        self.t = .27
        self.m = .35
        self.b = 1 - self.t - self.m
        self.pad = 4

    def init(self, controller):
        self.data_panel = DataPanel(controller, pos_hint={'x': 0., 'y': self.b + self.m}, size_hint=(self.l, self.t))
        self.feature_panel = FeaturePanel(controller, pos_hint={'x': self.l, 'y': self.b + self.m},
                                          size_hint=(self.r, self.t))
        self.predictor_panel = PredictorPanel(controller, self.m / self.t, pos_hint={'x': 0., 'y': self.b},
                                              size_hint=(self.l, self.m))
        self.plot_panel = PlotPanel(controller, pos_hint={'x': self.l, 'y': self.b}, size_hint=(self.r, self.m))
        self.extractor_panel = ExtractorPanel(controller, self.b / self.t, pos_hint={'x': 0., 'y': 0},
                                              size_hint=(self.l, self.b))
        self.theory_panel = TheoryPanel(controller, pos_hint={'x': self.l, 'y': 0}, size_hint=(self.r, self.b))

        self.data_panel.init()
        self.feature_panel.init()
        self.predictor_panel.init()
        self.plot_panel.init()
        self.extractor_panel.init()
        self.theory_panel.init()

        self.add_widget(self.data_panel)
        self.add_widget(self.feature_panel)
        self.add_widget(self.predictor_panel)
        self.add_widget(self.plot_panel)
        self.add_widget(self.extractor_panel)
        self.add_widget(self.theory_panel)
        self.bind(size=self.repaint)

    def draw(self):
        with self.canvas.before:
            Color(1, 1, 1)
            Rectangle(pos=[0, 0], size=self.size)
            Color(0, 0, 0)
            RoundedRectangle(pos=[self.pad, self.pad],
                             size=[self.width - self.pad * 2, self.height * self.b - self.pad])
            RoundedRectangle(pos=[self.pad, self.height * self.b + self.pad],
                             size=[self.width * self.l - self.pad, self.height * self.m - self.pad])
            RoundedRectangle(pos=[self.width * self.l + self.pad, self.height * self.b + self.pad],
                             size=[self.width * self.r - self.pad * 2, self.height * self.m - self.pad])
            RoundedRectangle(pos=[self.pad, self.height * (self.b + self.m) + self.pad],
                             size=[self.width - self.pad * 2, self.height * self.t - self.pad])
            Color(1, 1, 1)
            RoundedRectangle(pos=[self.width * (self.l + self.r / 2) - self.pad * 5, self.height * .95 - self.pad * 4],
                             size=[self.width * self.r / 2, self.height * .05])
            for i in range(3):
                RoundedRectangle(pos=[self.width * (self.l + self.r * i / 3) + self.pad * (2 - i),
                                      self.height * self.b + self.pad * 2],
                                 size=[(self.width * self.r - self.pad * 6) / 3, self.height * self.m - self.pad * 3])
            Line(points=[self.width * self.l * .575, 0, self.width * self.l * .575, self.height])
            Line(points=[self.width * self.l + self.pad / 2, 0, self.width * self.l + self.pad / 2, self.height])

    def repaint(self, widget, size):
        self.draw()


class RedLayout(RelativeLayout):

    def __init__(self, **kw):
        super().__init__(**kw)
        with self.canvas:
            Color(1, 0, 0)
            Rectangle(pos=self.pos, size=self.size)

    def init(self):
        pass




from pathlib import Path
from typing import Iterable, Dict, TextIO
import csv

PATH = Path(__file__).parents[0]

# TODO: add methods for file management,
#  add method for saving file onnx from a binary string


def get_test_path(filename: str) -> Path:
    return PATH / f"{filename}.csv"


def open_test(filename: str) -> TextIO:
    return open(get_test_path(filename))


def test_cases(filename: str) -> Iterable[Dict]:
    with open_test(filename) as file:
        return [row for row in csv.DictReader(file, delimiter=';', quotechar='"')]
